from abc import ABC
from typing import TYPE_CHECKING, Generic, Optional, TypeVar, Union

from pydantic import BaseModel, Field

from pydango.orm.encoders import jsonable_encoder
from pydango.orm.models import BaseArangoModel, CollectionConfig, CollectionType
from pydango.query.consts import FROM, TO

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny

TEdge = TypeVar("TEdge", bound="EdgeModel")


class EdgeCollectionConfig(CollectionConfig):
    type = CollectionType.EDGE


class EdgeModel(BaseArangoModel, ABC):
    from_: Optional[str] = Field(None, alias=FROM)
    to: Optional[Union[str]] = Field(None, alias=TO)

    class Collection(EdgeCollectionConfig):
        pass

    def save_dict(self) -> "DictStrAny":
        exclude: set[Union[int, str]] = set()
        for key in ["from_", "to"]:
            if self.__getattribute__(key) is None:
                exclude.add(key)
        return jsonable_encoder(self, by_alias=True, exclude=exclude)
        # return self.dict(by_alias=True, exclude=exclude)


T = TypeVar("T", bound=BaseModel)


class EdgeData(BaseModel, ABC, Generic[T]):
    pass


class EdgeDict(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{item}'")

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        try:
            del self[item]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{item}'")



from .connection.session import PydangoSession
from .orm import (
    EdgeCollectionConfig,
    EdgeModel,
    ORMQuery,
    Relation,
    VertexCollectionConfig,
    VertexModel,
)
from .query import AQLQuery
from .query.operations import TraversalDirection

__version__ = "0.2.0"
__all__ = [
    "PydangoSession",
    "VertexModel",
    "EdgeModel",
    "EdgeCollectionConfig",
    "VertexCollectionConfig",
    "ORMQuery",
    "AQLQuery",
    "Relation",
    "TraversalDirection",
]



from abc import ABCMeta, abstractmethod
from enum import Enum, IntEnum
from functools import partial
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    ForwardRef,
    Generic,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

from pydantic import BaseConfig, ConfigError, Field
from pydantic.fields import SHAPE_SINGLETON, ModelField, PrivateAttr, Undefined
from pydantic.main import BaseModel, ModelMetaclass
from pydantic.typing import resolve_annotations

from pydango.connection.consts import PYDANGO_SESSION_KEY
from pydango.indexes import Indexes
from pydango.orm.consts import EDGES
from pydango.orm.encoders import jsonable_encoder
from pydango.orm.models.fields import (
    ModelFieldExpression,
    RelationModelField,
    get_pydango_field,
)
from pydango.orm.models.relations import Relationship
from pydango.orm.models.sentinel import LazyFetch
from pydango.orm.models.shapes import LIST_SHAPES
from pydango.orm.utils import evaluate_forward_ref
from pydango.query.consts import ID, KEY, REV
from pydango.query.expressions import FieldExpression, ObjectExpression

if TYPE_CHECKING:
    from pydantic.main import GetterDict, Model
    from pydantic.typing import AbstractSet, DictStrAny, MappingIntStrAny

    from pydango.connection.session import PydangoSession
    from pydango.orm.models.types import RelationshipFields, Relationships
    from pydango.orm.models.vertex import TVertexModel

ArangoModel = TypeVar("ArangoModel", bound="BaseArangoModel")
FieldType = TypeVar("FieldType")
OPERATIONAL_FIELDS = {"key", "id", "rev"}


class LinkTypes(str, Enum):
    DIRECT = "DIRECT"
    OPTIONAL_DIRECT = "OPTIONAL_DIRECT"
    LIST = "LIST"
    OPTIONAL_LIST = "OPTIONAL_LIST"
    EDGE = "EDGE"
    OPTIONAL_EDGE = "OPTIONAL_EDGE"
    EDGE_LIST = "EDGE_LIST"
    OPTIONAL_EDGE_LIST = "OPTIONAL_EDGE_LIST"


EDGE_TYPES = (
    LinkTypes.EDGE,
    LinkTypes.OPTIONAL_EDGE,
    LinkTypes.OPTIONAL_EDGE_LIST,
    LinkTypes.EDGE_LIST,
)

LIST_TYPES = (
    LinkTypes.EDGE_LIST,
    LinkTypes.OPTIONAL_EDGE_LIST,
    LinkTypes.LIST,
    LinkTypes.OPTIONAL_LIST,
)

SINGLETON_TYPES = (
    LinkTypes.EDGE,
    LinkTypes.DIRECT,
    LinkTypes.OPTIONAL_EDGE,
    LinkTypes.OPTIONAL_DIRECT,
)


def get_relation(field_name: str, annotation: Any, value: Any, _: Type["BaseConfig"]) -> Optional["Relationship"]:
    if get_origin(annotation) is not Annotated:
        return None

    args = get_args(annotation)
    relation_infos = [arg for arg in args[1:] if arg is Relation or get_origin(arg) is Relation]
    if len(relation_infos) > 1:
        raise ValueError(f"cannot specify multiple `Annotated` `Field`s for {field_name!r}")
    relation_info = next(iter(relation_infos), None)
    via_model = get_args(relation_info)[0] if relation_info else None
    field = ModelField.infer(
        name=field_name,
        value=value,
        annotation=annotation,
        class_validators=None,
        config=BaseConfig,
    )
    if field.shape in LIST_SHAPES:
        if field.sub_fields:
            link_model = field.sub_fields[0].type_

        if field.allow_none is True:
            link_type = via_model and LinkTypes.OPTIONAL_EDGE_LIST or LinkTypes.OPTIONAL_LIST
        else:
            link_type = via_model and LinkTypes.EDGE_LIST or LinkTypes.LIST

    elif field.shape == SHAPE_SINGLETON:
        link_model = field.type_
        if field.allow_none is True:
            link_type = via_model and LinkTypes.OPTIONAL_EDGE or LinkTypes.OPTIONAL_DIRECT
        else:
            link_type = via_model and LinkTypes.EDGE or LinkTypes.DIRECT
    else:
        raise AssertionError()

    return Relationship(
        field=field,
        link_model=link_model,
        link_type=link_type,
        via_model=via_model,
    )


class Relation(Generic[ArangoModel]):
    def __init__(self, *args, **kwargs):
        pass


class LazyProxy(Generic[ArangoModel]):
    _initialized: bool = False
    __instance__: Union[ArangoModel, "LazyFetch"]

    def __init__(
        self, instance: Union[ArangoModel, "LazyFetch"], field, parent: ArangoModel, session: Optional["PydangoSession"]
    ):
        self.parent = parent
        self.session = session
        self._relation_field = field
        if not isinstance(instance, LazyFetch):
            self._initialized = True

        self.__instance__ = instance

    def __getattr__(self, item):
        if item in getattr(self, "__dict__"):
            return getattr(self, item)

        if isinstance(self.__instance__, list):
            if item in ["dict"]:
                return partial(jsonable_encoder, obj=self.__instance__)

        attr = getattr(self.__instance__, item, None)
        if attr:
            return attr
        else:
            return getattr(self._relation_field.type_, item)

    def __repr__(self):
        return repr(self.__instance__)

    def __getitem__(self, item):
        if self:
            return self.__instance__[item]
        raise AttributeError(
            f"you are attempting to access {self._relation_field.field.type_.__name__} via"
            f" {self._relation_field.field.name} which is not initialized yet, use fetch"
        )

    def __bool__(self):
        return self._initialized and bool(self.__instance__)

    async def fetch(
        self,
    ):
        if not self.session:
            raise "Kusomo"

        model = await self.session.get(
            self.parent.__class__,
            # self._relation_field.field.type_,
            self.parent.key,
            fetch_edges={self._relation_field.via_model.Collection.name},
            depth=range(1, 1),
        )
        model = getattr(model, self._relation_field.field.name)
        setattr(self.parent, self._relation_field.field.name, model)
        self.__instance__ = model
        self._initialized = True
        return model

    def compile(self, query_ref):
        return ObjectExpression(self.dict()).compile(query_ref)

    def dict(self, *args, by_alias=True, **kwargs):
        return jsonable_encoder(self.__instance__, by_alias=by_alias, *args, **kwargs)


class DocFieldDescriptor(Generic[FieldType]):
    def __init__(self, field: ModelField, relation: Optional[Relationship] = None):
        self.relation = relation
        self.field = field
        self._proxy: Optional[LazyProxy] = None

    def __set__(self, instance, value: FieldType):
        raise AssertionError()
        # instance.__dict__[self.name] = LazyProxy(value)

    def __get__(
        self, instance: Optional[ArangoModel], owner: Type["TVertexModel"]
    ) -> Union[LazyProxy[ArangoModel], ModelFieldExpression, FieldType, None]:
        if not instance and self.field.name in owner.__fields__.keys():
            return ModelFieldExpression(self.field.name, owner)
        if instance:
            field_value = instance.__dict__.get(self.field.name)
            if self.field.name in instance.__fields_set__:
                return field_value

            if self._proxy:
                return self._proxy

            if self.relation:
                session = getattr(instance, PYDANGO_SESSION_KEY, None)
                self._proxy = LazyProxy(field_value, self.relation, instance, session)
                return self._proxy  # type: ignore[valid-type]
        if not instance:
            raise ValueError("something happened open an issue :(")
        return None

    def __set_name__(self, owner, name):
        self.name = name


class ArangoModelMeta(ModelMetaclass, ABCMeta):
    def __new__(mcs, name: str, bases: tuple[Type], namespace: dict, **kwargs: Any):
        parents = [b for b in bases if isinstance(b, mcs)]
        if not parents or BaseArangoModel in parents:
            skipped_cls: BaseArangoModel = super().__new__(mcs, name, bases, namespace, **kwargs)
            skipped_cls.__relationships__ = {}
            skipped_cls.__relationships_fields__ = {}
            return skipped_cls

        _relationships, original_annotations = ArangoModelMeta.get_relations_from_namespace(namespace)

        dict_used = {
            **namespace,
            "__weakref__": None,
            "__annotations__": original_annotations,
            "__relationships__": _relationships,
        }

        new_cls: BaseArangoModel = super().__new__(mcs, name, bases, dict_used, **kwargs)

        __relationship_fields__ = ArangoModelMeta.set_field_descriptors(_relationships, new_cls)

        new_cls.__relationships__ = _relationships
        new_cls.__relationships_fields__ = __relationship_fields__
        new_cls.__annotations__ = {
            # **relationship_annotations,
            **original_annotations,
            **new_cls.__annotations__,
        }

        return new_cls

    @staticmethod
    def get_relations_from_namespace(namespace: dict[str, Any]) -> tuple["Relationships", dict[str, Any]]:
        _relationships: dict[str, Relationship] = {}
        original_annotations = resolve_annotations(
            namespace.get("__annotations__", {}), namespace.get("__module__", None)
        )
        for k, v in original_annotations.items():
            relation = get_relation(k, v, namespace.get(k, Undefined), BaseConfig)
            if relation:
                _relationships[k] = relation
                # original_annotations[k] = Union[original_annotations[k]]
        return _relationships, original_annotations

    @staticmethod
    def set_field_descriptors(_relationships, new_cls):
        __relationship_fields__ = {}
        for field_name, model_field in [(x, y) for x, y in new_cls.__fields__.items() if x != EDGES]:
            if field_name in _relationships:
                pydango_field = get_pydango_field(model_field, RelationModelField)
                relationship = _relationships[field_name]
                relationship.field = pydango_field
                __relationship_fields__[field_name] = pydango_field
                new_cls.__fields__[field_name] = pydango_field

                type_ = cast(ModelField, pydango_field).type_
                setattr(
                    new_cls,
                    field_name,
                    DocFieldDescriptor[type_](pydango_field, relationship),  # type: ignore[valid-type]
                )

                field_annotation = {field_name: DocFieldDescriptor[type_]}  # type: ignore[valid-type]
                new_cls.__annotations__.update(field_annotation)
            else:
                setattr(
                    new_cls,
                    field_name,
                    DocFieldDescriptor[model_field.type_](model_field),  # type: ignore[name-defined]
                )
        return __relationship_fields__

    # def __hash__(self):
    #     return hash(self.Collection.name)


class CollectionType(IntEnum):
    NODE = 2
    EDGE = 3


class CollectionConfig:
    name: str
    type: CollectionType
    sync: Optional[bool] = False
    system: Optional[bool] = False
    key_generator: Literal["traditional", "autoincrement"] = "traditional"
    user_keys: Optional[bool] = False
    key_increment: Optional[int]
    key_offset: Optional[int]
    shard_fields: Optional[list[str]]
    shard_count: Optional[int]
    replication_factor: Optional[int]
    shard_like: Optional[str]  # enterprise only
    sync_replication: Optional[bool]
    enforce_replication_factor: Optional[bool]
    sharding_strategy: Literal["community-compat", "enterprise-smart-edge-compat", "enterprise-smart-edge"]
    smart_join_attribute: Optional[str]  # enterprise only
    write_concern: Optional[int]

    sync_schema: Optional[bool] = False
    indexes: Sequence[Indexes] = []


class BaseArangoModel(BaseModel, metaclass=ArangoModelMeta):
    id: Optional[str] = Field(None, alias=ID)
    key: Optional[str] = Field(None, alias=KEY)
    rev: Optional[str] = Field(None, alias=REV)

    __session__: Optional["PydangoSession"] = PrivateAttr()

    if TYPE_CHECKING:
        __relationships__: Relationships = {}
        __relationships_fields__: RelationshipFields = {}

    class Config(BaseConfig):
        arbitrary_types_allowed = True
        orm_mode = True
        # getter_dict = dict
        allow_population_by_field_name = True

    class Collection(CollectionConfig): ...

    def __init__(self, **data: Any):
        super().__init__(**data)
        object.__setattr__(self, PYDANGO_SESSION_KEY, data.get(PYDANGO_SESSION_KEY))

    @classmethod
    def _decompose_class(cls: Type["Model"], obj: Any) -> Union["GetterDict", dict]:  # type: ignore[override]
        if isinstance(obj, dict):
            return obj
        decompose_class = super()._decompose_class(obj)
        return decompose_class

    def _calculate_keys(
        self,
        include: Optional["MappingIntStrAny"],
        exclude: Optional["MappingIntStrAny"],
        exclude_unset: bool,
        update: Optional["DictStrAny"] = None,
    ) -> Optional["AbstractSet[str]"]:
        field_set = self.__fields_set__.copy()
        keys = self.__dict__.keys()
        unset = keys - field_set
        if not exclude_unset:
            _exclude = cast(Mapping, {field: True for field in unset if field in OPERATIONAL_FIELDS})
            if not exclude:
                exclude = _exclude
            else:
                exclude.update(_exclude)  # type: ignore[attr-defined]

        return super()._calculate_keys(include, exclude, exclude_unset, update)

    @classmethod
    def from_orm(cls: Type[ArangoModel], obj: Any, *, session=None) -> ArangoModel:
        obj[PYDANGO_SESSION_KEY] = session
        for field_name, field in cls.__relationships_fields__.items():
            exists_in_orm = field_name in obj and obj.get(field_name, None)
            if exists_in_orm:
                if isinstance(exists_in_orm, list):
                    for i, v in enumerate(exists_in_orm):
                        exists_in_orm[i][PYDANGO_SESSION_KEY] = session
                else:
                    exists_in_orm[PYDANGO_SESSION_KEY] = session

                obj[field_name] = exists_in_orm

                continue
            if field.required:
                obj[field_name] = LazyFetch(session, obj["_id"])
            else:
                print("field not set", field_name)

        try:
            obj = cast(Type[ArangoModel], super()).from_orm(obj)
        except ConfigError as e:
            raise e

        # for field_name, field in cls.__relationships_fields__.items():
        #     setattr( getattr(obj,field_name),'__dali_session__',session)
        # object_setattr(obj, DALI_SESSION_KW, session)
        return obj

    # @classmethod
    # def validate(cls: Type['Model'], value: Any) -> 'Model':
    #     return cls.from_orm(value)

    @classmethod
    def update_forward_refs(cls, **localns: Any) -> None:
        super().update_forward_refs(**localns)
        for name in cls.__relationships_fields__.keys():
            cls.__relationships_fields__[name] = cast(RelationModelField, cls.__fields__[name])
            relation = cls.__relationships__[name]
            relation.field = cls.__fields__[name]
            relation.link_model = cls.__fields__[name].type_
            if isinstance(relation.via_model, ForwardRef):
                relation.via_model = evaluate_forward_ref(cls, relation.via_model, **localns)

            if isinstance(relation.link_model, ForwardRef):
                relation.link_model = evaluate_forward_ref(cls, relation.link_model, **localns)

        # for k in cls.__edge_to_field_mapping__.copy():
        #     if isinstance(k, ForwardRef):
        #         funcs = cls.__edge_to_field_mapping__.pop(k)
        #         new_k = evaluate_forward_ref(cls, k, **localns)
        #         if new_k in cls.__edge_to_field_mapping__:
        #             cls.__edge_to_field_mapping__[new_k.Collection.name].extend(funcs)
        #         else:
        #             cls.__edge_to_field_mapping__[new_k.Collection.name] = funcs

    @abstractmethod
    def save_dict(self) -> "DictStrAny": ...


class Aliased(Generic[ArangoModel]):
    def __init__(self, entity: ArangoModel, alias=None):
        self.entity: ArangoModel = entity
        self.alias = alias

    def __getattr__(self, item):
        attr = getattr(self.entity, item)
        if isinstance(attr, FieldExpression):
            attr.parent = self

        return attr

    def __str__(self):
        return str(self.alias or "")

    def __repr__(self):
        return f"<Aliased: {self.entity.__name__}, {id(self)}>"

    # def compile(self, query_ref: "ORMQuery") -> str:
    #     return query_ref.orm_bound_vars[self].compile(query_ref)



from typing import TYPE_CHECKING

from pydango.orm.models.fields import ModelFieldExpression

if TYPE_CHECKING:
    from pydango.orm.models.base import ArangoModel


def save_dict(model: "ArangoModel"):
    return model.save_dict()


def convert_edge_data_to_valid_kwargs(edge_dict):
    for i in edge_dict.copy():
        if isinstance(i, ModelFieldExpression):
            edge_dict[i.field] = edge_dict.pop(i)



import datetime
import sys
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from functools import lru_cache
from typing import Any, Mapping, Optional, Sequence, Type, Union, cast

from pydango.query.consts import DYNAMIC_ALIAS

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


class SortDirection(str, Enum):
    ASC = "ASC"
    DESC = "DESC"


class ReturnableExpression(ABC):
    """
    Base class for returnable expressions
    """


class Expression(ABC):
    @abstractmethod
    def compile(self, query_ref: "QueryExpression") -> str:
        raise NotImplementedError

    def __deepcopy__(self, memodict=None):
        return self


class BindableExpression(Expression):
    def __init__(self, value: Any) -> None:
        self.value = value

    def compile(self, query_ref: "QueryExpression") -> str:
        # Mypy: Argument 1 to "bind_parameter" of "AQLQuery"
        # has incompatible type "expressions.LiteralExpression";
        # expected "Union[pydango.query.expressions.LiteralExpression, FigurativeExpression]"  [arg-type]
        return query_ref.bind_parameter(self)  # type: ignore[arg-type]


class LiteralExpression(BindableExpression):
    def __repr__(self):
        return "?"


class VariableExpression(Expression, ReturnableExpression):
    def __init__(self, var_name: Optional[str] = None):
        self.var_name = var_name

    def compile(self, query_ref: "QueryExpression") -> str:
        if self.var_name is None:
            self.var_name = query_ref.bind_variable()

        return self.var_name

    def __repr__(self):
        return self.var_name or DYNAMIC_ALIAS

    def __getattr__(self, item):
        if item in self.__dict__:
            return getattr(self, item)
        else:
            return FieldExpression(item, self)

    def __getitem__(self, item):
        return FieldExpression(item, self)


class ModificationVariable(VariableExpression, ABC):
    _keyword: str

    def compile(self, query_ref: "QueryExpression") -> str:
        if not query_ref.__is_modification_query__:
            raise AssertionError(f"no modification operation defined to use {self._keyword} keyword")

        return self._keyword

    def __repr__(self):
        return self._keyword

    def __getattr__(self, item):
        if item in self.__dict__:
            return getattr(self, item)
        else:
            return FieldExpression(item, self)


class NEW(ModificationVariable):
    _keyword = "NEW"


class OLD(ModificationVariable):
    _keyword = "OLD"


class FieldExpression(Expression, ReturnableExpression):
    """
    Expression class for field access of objects and documents
    """

    def __init__(
        self, field: Union[str, Expression], parent: Union[VariableExpression, "FieldExpression", None] = None
    ):
        self.parent = parent
        self.field = field

    def compile(self, query_ref: "QueryExpression") -> str:
        if self.parent is None:
            raise ValueError("field parent must be populated")
        if isinstance(self.field, Expression):
            return f"{self.parent.compile(query_ref)}[{self.field.compile(query_ref)}]"
        if isinstance(self.field, int):
            return f"{self.parent.compile(query_ref)}[{self.field}]"
        else:
            return f"{self.parent.compile(query_ref)}.{self.field}"

    def __getattr__(self, item):
        return FieldExpression(item, self)

    def __repr__(self):
        return f"{repr(self.parent)}.{str(self.field)}"

    def __pos__(self) -> "SortExpression":
        return SortExpression(self, SortDirection.ASC)

    def __neg__(self) -> "SortExpression":
        return SortExpression(self, SortDirection.DESC)

    def __eq__(self, other) -> "ConditionExpression":  # type: ignore[override]
        return cast(ConditionExpression, _set_operator(self, "==", other, ConditionExpression))

    def __ge__(self, other) -> "ConditionExpression":
        return cast(ConditionExpression, _set_operator(self, ">=", other, ConditionExpression))

    def __gt__(self, other) -> "ConditionExpression":
        return cast(ConditionExpression, _set_operator(self, ">", other, ConditionExpression))

    def __le__(self, other) -> "ConditionExpression":
        return cast(ConditionExpression, _set_operator(self, "<=", other, ConditionExpression))

    def __lt__(self, other) -> "ConditionExpression":
        return cast(ConditionExpression, _set_operator(self, "<", other, ConditionExpression))

    def __mul__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "*", other, ArithmeticExpression))

    def __sub__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "-", other, ArithmeticExpression))

    def __add__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "+", other, ArithmeticExpression))

    def __truediv__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "/", other, ArithmeticExpression))

    def __mod__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "%", other, ArithmeticExpression))


class IteratorExpression(VariableExpression):
    def __repr__(self):
        return f"<Iterator: {str(self.var_name or DYNAMIC_ALIAS)} {id(self)}>"

    def __getattr__(self, item):
        if item in self.__dict__:
            return getattr(self, item)
        else:
            return FieldExpression(item, self)

    def __gt__(self, other):
        return _set_operator(self, ">", other, ConditionExpression)

    def __ge__(self, other) -> "ConditionExpression":
        return cast(ConditionExpression, _set_operator(self, ">=", other, ConditionExpression))

    def __lt__(self, other):
        return _set_operator(self, "<", other, ConditionExpression)

    def __le__(self, other) -> "ConditionExpression":
        return cast(ConditionExpression, _set_operator(self, "<=", other, ConditionExpression))

    def __eq__(self, other):
        return cast(ConditionExpression, _set_operator(self, "==", other, ConditionExpression))

    def __ne__(self, other):
        return cast(ConditionExpression, _set_operator(self, "!=", other, ConditionExpression))

    def __mul__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "*", other, ArithmeticExpression))

    def __sub__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "-", other, ArithmeticExpression))

    def __add__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "+", other, ArithmeticExpression))

    def __truediv__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "/", other, ArithmeticExpression))

    def __mod__(self, other) -> "ArithmeticExpression":
        return cast(ArithmeticExpression, _set_operator(self, "%", other, ArithmeticExpression))


class IterableExpression(Expression, ReturnableExpression, ABC):
    """
    Base class for iterable expressions
    """

    # todo: move this to a more relevant place
    def __init__(self, iterator: Optional[Union[IteratorExpression, str]] = None):
        if iterator is None:
            self.iterator = IteratorExpression()
        elif isinstance(iterator, str):
            self.iterator = IteratorExpression(iterator)
        else:
            self.iterator = iterator


class QueryExpression(Expression, ABC):
    parent: Optional["QueryExpression"] = None
    sep = " "
    __used_vars__: set[str]
    __is_modification_query__: bool

    @abstractmethod
    def bind_variable(self) -> str: ...

    @abstractmethod
    def bind_parameter(self, parameter: "BindableExpression", override_var_name: Optional[str] = None) -> str: ...


class RangeExpression(IterableExpression):
    def __init__(self, start, end):
        super().__init__()
        self.end = end
        self.start = start

    def compile(self, query_ref: "QueryExpression"):
        if isinstance(self.start, Expression):
            start = self.start.compile(query_ref)
        else:
            start = self.start
        if isinstance(self.end, Expression):
            end = self.end.compile(query_ref)
        else:
            end = self.end

        return f"{start}..{end}"

    def __repr__(self):
        return f"{self.start}..{self.end}"


class AssignmentExpression(Expression):
    def __init__(self, variable: VariableExpression, expression: Expression):
        self.variable = variable
        self.expression = expression

    def compile(self, query_ref: "QueryExpression"):
        expression_compile = self.expression.compile(query_ref)
        if isinstance(self.expression, QueryExpression):
            self.expression.parent = query_ref
            expression_compile = f"({expression_compile})"
        return f"{self.variable.compile(query_ref)} = {expression_compile}"

    def __repr__(self):
        expression_repr = repr(self.expression)
        if isinstance(self.expression, QueryExpression):
            expression_repr = f"({expression_repr})"
        return f"{repr(self.variable)} = {expression_repr}"


class UnaryExpression(Expression, ReturnableExpression):
    """
    Expression class for unary operations
    """

    def __init__(self, operand: Expression, operator: str):
        self.operand = operand
        self.operator = operator

    def compile(self, query_ref: "QueryExpression") -> str:
        return f"{self.operator}({self.operand.compile(query_ref)})"


class BinaryExpression(Expression, ReturnableExpression):
    """
    Expression class for binary operations
    """

    def __init__(self, op: str, left: Expression, right: Expression) -> None:
        self.left = left
        self.op = op
        self.right = right

    @lru_cache(maxsize=1)
    def compile(self, query_ref: "QueryExpression") -> str:
        left_compile = self.left.compile(query_ref)
        right_compile = self.right.compile(query_ref)
        return f"{left_compile} {self.op} {right_compile}"

    def __repr__(self):
        return f"{self.left} {self.op} {self.right}"


class LogicalExpression(Expression, ReturnableExpression, ABC):
    """
    Base class for logical expressions
    """


class UnaryLogicalExpression(UnaryExpression, LogicalExpression):
    pass


class ConditionExpression(BinaryExpression, LogicalExpression):
    """
    valid operators >, >=, ==, !=, <, <=
    """

    def __repr__(self):
        return f"{str(self.left)} {self.op} {str(self.right)}"

    def __and__(self, other):
        return AndExpression(self, other)

    def __or__(self, other):
        return OrExpression(self, other)


class BinaryLogicalExpression(ConditionExpression):
    """
    Expression class for binary logical operations
    valid operators &&, AND, ||, OR
    """

    @lru_cache(maxsize=1)
    def compile(self, query_ref: "QueryExpression") -> str:
        return f"({self.left.compile(query_ref)} {self.op} {self.right.compile(query_ref)})"


class AndExpression(BinaryLogicalExpression):
    """
    Expression class for binary logical operations
    valid operators &&, AND, ||, OR
    """

    def __init__(self, left: Expression, right: Expression):
        super().__init__("&&", left, right)


class OrExpression(BinaryLogicalExpression):
    """
    Expression class for binary logical operations
    valid operators &&, AND, ||, OR
    """

    def __init__(self, left: Expression, right: Expression):
        super().__init__("||", left, right)


class In(ConditionExpression):
    def __init__(self, obj: LiteralExpression, iterable: IterableExpression):
        super().__init__("IN", obj, iterable)


class BaseArithmeticExpression(Expression, ReturnableExpression, ABC):
    """
    Base class for arithmetic expressions
    """


class UnaryArithmeticExpression(UnaryExpression, BaseArithmeticExpression): ...


class ArithmeticExpression(BinaryExpression, BaseArithmeticExpression):
    def __lt__(self, other):
        return ConditionExpression("<", self, other)

    def __le__(self, other):
        return ConditionExpression("<=", self, other)

    def __gt__(self, other):
        return ConditionExpression(">", self, other)

    def __ge__(self, other):
        return ConditionExpression(">=", self, other)

    def __eq__(self, other):
        return ConditionExpression("==", self, other)

    def __hash__(self):
        return id(self)


class ReturnableIterableExpression(IterableExpression, ReturnableExpression, ABC):
    def __init__(self, iterator):
        super().__init__(iterator)


class SubQueryExpression(Expression, ReturnableExpression):
    def __init__(self, query: QueryExpression):
        self.query = query
        self.query.sep = " "

    def __repr__(self):
        if self.query.sep == "\n":
            return f"({repr(self.query)})"
        return f"({repr(self.query)})"

    def compile(self, query_ref) -> str:
        # self.query.sep = "\n"
        if self.query.sep == "\n":
            return f"({self.query.compile(query_ref)})"
        return f"({self.query.compile(query_ref)})"


class ScalarSubQuery(SubQueryExpression): ...


class VectorSubQueryExpression(SubQueryExpression, IterableExpression):
    def __init__(self, query: "QueryExpression"):
        super().__init__(query)
        super(SubQueryExpression, self).__init__(IteratorExpression())

        if self.query.sep == "\n":
            self.query.sep = f"{self.query.sep}\t"


class CollectionExpression(IterableExpression):
    def __init__(self, collection_name: str, iterator: Optional[Union[str, IteratorExpression]] = None):
        super().__init__(iterator)
        self.collection_name = collection_name

    def __repr__(self):
        return f"<CollectionExpression: {str(self.collection_name)}>"

    def __getattr__(self, item) -> Union[Any, FieldExpression]:
        if item in self.__dict__:
            return getattr(self, item)
        else:
            return getattr(self.iterator, item)

    def compile(self, *args, **kwargs) -> str:
        return f"`{self.collection_name}`"


class FigurativeExpression(BindableExpression, ReturnableExpression, ABC):
    pass


ListItems: TypeAlias = Union[
    QueryExpression, LiteralExpression, FigurativeExpression, Mapping, Sequence, int, float, str, bool
]
ListValues: TypeAlias = Union[
    tuple[
        ListItems,
        ...,
    ],
    list[ListItems],
]


class ListExpression(
    BindableExpression,
    IterableExpression,
):
    def __init__(self, value: ListValues, iterator: Optional[Union[IteratorExpression, str]] = None, brackets=True):
        if isinstance(value, list):
            value = tuple(value)

        super().__init__(value)
        super(BindableExpression, self).__init__(iterator)
        self._copy: list[Expression] = []
        self._need_compile = False
        self._brackets = brackets
        for i in self.value:
            if isinstance(i, QueryExpression):
                self._copy.append(SubQueryExpression(i))
                self._need_compile = True
            elif isinstance(i, (VariableExpression, FieldExpression)):
                self._copy.append(i)
                self._need_compile = True
            elif isinstance(i, Expression):
                self._copy.append(i)
            elif isinstance(i, list):  # todo review this
                self._copy.append(ListExpression(i))
                self._need_compile = True
            elif isinstance(i, dict):
                expression = ObjectExpression(i)
                self._copy.append(expression)
                self._need_compile = not expression.__all_literals__ or self._need_compile
            else:
                self._copy.append(LiteralExpression(i))

    def compile(self, query_ref: "QueryExpression", **kwargs) -> str:
        if self._need_compile:
            result = []
            for i in self._copy:
                if isinstance(i, SubQueryExpression):
                    i.query.parent = query_ref
                result.append(i.compile(query_ref))
            if self._brackets:
                return f'[{", ".join(result)}]'
            else:
                return ", ".join(result)

        return super().compile(query_ref)

    def __repr__(self):
        if self._need_compile:
            result = []
            for i in self._copy:
                result.append(repr(i))
            return f'[{", ".join(result)}]'

        return "?"


ObjectKeys = Union[str, VariableExpression, CollectionExpression, IteratorExpression]

ObjectValues = Union[
    LiteralExpression, FigurativeExpression, FieldExpression, Mapping, Sequence, int, float, str, bool, None
]

ObjectParams: TypeAlias = Mapping[
    ObjectKeys,
    ObjectValues,
]


class ObjectExpression(BindableExpression, ReturnableExpression):
    def __init__(self, value: ObjectParams, parent: Optional[Union[VariableExpression, CollectionExpression]] = None):
        super().__init__(value)
        self.parent = parent
        self._bind = {}
        self.__all_literals__ = True
        if isinstance(value, dict):
            self.value = deepcopy(value)
            for field, mapped_field in self.value.items():
                if isinstance(mapped_field, list):
                    self.value[field] = ListExpression(mapped_field)
                    self.__all_literals__ = self.__all_literals__ or not self.value[field]._need_compile
                elif isinstance(mapped_field, dict):
                    self.value[field] = ObjectExpression(mapped_field, self.parent)
                    self.__all_literals__ = self.__all_literals__ or self.value[field].__all_literals__
                elif isinstance(mapped_field, QueryExpression):
                    subquery = SubQueryExpression(mapped_field)
                    self.value[field] = subquery
                    self._bind[mapped_field] = subquery
                    self.__all_literals__ = False
                elif isinstance(mapped_field, (VariableExpression, FieldExpression)):
                    self.__all_literals__ = False
                elif (
                    isinstance(mapped_field, (int, float, str, bool, datetime.datetime, datetime.date))
                    or mapped_field is None
                ):
                    self.value[field] = LiteralExpression(mapped_field)

    def compile(self, query_ref: "QueryExpression") -> str:
        for bind in self._bind.values():
            bind.query.parent = query_ref

        pairs = []

        if isinstance(self.value, dict):
            for field, mapped_field in self.value.items():
                if isinstance(field, Expression):
                    field = field.compile(query_ref)
                pairs.append(f"{field}: {mapped_field.compile(query_ref)}")

        return f"{{{', '.join(pairs)}}}"

    def __repr__(self):
        pairs = []
        # if isinstance(self.value, list):
        #     for field in self.value:
        #         pairs.append(f"{field}: {self.parent}.{field}")
        if isinstance(self.value, dict):
            for field, mapped_field in self.value.items():
                pairs.append(f"{field}: {mapped_field}")
        return f"{{{', '.join(pairs)}}}"


class BaseAQLVariableExpressionMixin(Expression):
    def __init__(self, value: str):
        self.value = value


# class AQLVariableExpression(BaseAQLVariableExpressionMixin):
#     def __init__(self, var_name: str):
#         super().__init__(var_name)
#
#     def compile(self, *args, **kwargs) -> str:
#         return f"@{super().compile(*args, **kwargs)}"


class AQLCollectionVariableExpression(VariableExpression):
    def __init__(self, value: str):
        super().__init__(value)

    def compile(self, *args, **kwargs) -> str:
        return f"@@{super().compile(*args, **kwargs)}"


def _set_operator(self, operator, other, cls: Type[BinaryExpression]) -> BinaryExpression:
    if not isinstance(other, Expression):
        other = LiteralExpression(other)
    return cls(operator, self, other)


class SortExpression(Expression):
    def __init__(self, field: FieldExpression, direction: Optional[SortDirection] = None):
        self.field = field
        self.direction = direction

    def compile(self, query_ref: "QueryExpression") -> str:
        compiled = f"{self.field.compile(query_ref)}"
        if self.direction:
            compiled += f" {self.direction.value}"
        return compiled

    def __repr__(self):
        if hasattr(self, "field") and hasattr(self, "direction"):
            return f"{self.field} {self.direction}"
        return super().__repr__()

    def __invert__(self):
        if self.direction == SortDirection.ASC:
            self.direction = SortDirection.DESC
        else:
            self.direction = SortDirection.ASC


class DynamicFieldExpression(FieldExpression):
    def compile(self, query_ref: "QueryExpression") -> str:
        return f"[{super().compile(query_ref)}]"

    def __hash__(self):
        field = ""
        if self.parent:
            field += f"{self.parent}."
        field += f"{self.field}"
        return hash(field)



from .base import BaseArangoModel, CollectionConfig, CollectionType, Relation
from .edge import EdgeCollectionConfig, EdgeModel
from .vertex import VertexCollectionConfig, VertexModel

__all__ = [
    "BaseArangoModel",
    "EdgeModel",
    "VertexModel",
    "CollectionConfig",
    "CollectionType",
    "EdgeCollectionConfig",
    "VertexCollectionConfig",
    "Relation",
]



from typing import (
    TYPE_CHECKING,
    Any,
    ForwardRef,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic.fields import Field, ModelField
from pydantic.main import create_model
from pydantic.typing import evaluate_forwardref

from pydango.orm.consts import EDGES
from pydango.orm.encoders import jsonable_encoder
from pydango.orm.models import BaseArangoModel, CollectionConfig, CollectionType
from pydango.orm.models.base import LIST_TYPES, ArangoModelMeta, LinkTypes
from pydango.orm.models.edge import EdgeData, EdgeDict
from pydango.orm.models.types import EdgeFieldMapping, Relationships
from pydango.orm.models.utils import convert_edge_data_to_valid_kwargs
from pydango.orm.utils import evaluate_forward_ref, get_globals

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny

TVertexModel = TypeVar("TVertexModel", bound="VertexModel")
TEdges = TypeVar("TEdges", bound=EdgeData)


class VertexCollectionConfig(CollectionConfig):
    type = CollectionType.NODE


class VertexMeta(ArangoModelMeta):
    def __new__(mcs, name: str, bases: tuple[Type], namespace: dict, **kwargs: Any):
        parents = [b for b in bases if isinstance(b, mcs)]
        if not parents:
            return super().__new__(mcs, name, bases, namespace, **kwargs)
        _relationships, original_annotations = mcs.get_relations_from_namespace(namespace)
        __edge_to_field_mapping__, edge_annotation = mcs.build_edges_model(_relationships, bases, name, namespace)

        namespace["__edge_to_field_mapping__"] = __edge_to_field_mapping__
        namespace["__annotations__"][EDGES] = edge_annotation

        return super().__new__(mcs, name, bases, namespace, **kwargs)

    @staticmethod
    def build_edges_model(
        _relationships: Relationships, bases: tuple[Type[Any]], name: str, namespace: dict[str, Any]
    ) -> tuple[EdgeFieldMapping, ModelField]:
        if VertexModel in bases:
            edges_model = VertexMeta._build_model(_relationships, name)
            namespace[EDGES] = Field(None, exclude=True)
            edge_annotation = cast(Any, Optional[edges_model])
        else:
            namespace[EDGES] = Field(None, exclude=True)
            edge_annotation = cast(Any, None)

        __edge_to_field_mapping__ = VertexMeta._build_edge_to_field_mapping(_relationships)

        VertexMeta._validate_edges(__edge_to_field_mapping__, namespace)
        return __edge_to_field_mapping__, edge_annotation

    @staticmethod
    def _build_edge_to_field_mapping(relationships: Relationships) -> EdgeFieldMapping:
        __edge_to_field_mapping__: EdgeFieldMapping = {}
        for relation_field, relation_info in relationships.items():
            if not relation_info.via_model:
                continue
            if isinstance(relation_info.via_model, ForwardRef):
                __edge_to_field_mapping__.setdefault(relation_info.via_model, []).append(cast(str, relation_field))
            elif issubclass(relation_info.via_model, BaseArangoModel):
                __edge_to_field_mapping__.setdefault(relation_info.via_model.Collection.name, []).append(relation_field)
        return __edge_to_field_mapping__

    @staticmethod
    def _validate_edges(edge_to_field_mapping: EdgeFieldMapping, namespace: dict[str, Any]) -> None:
        errors: dict[Union[str, ForwardRef], list[str]] = {}
        items = edge_to_field_mapping.items()
        for coll_or_forward_ref, fields in items:
            if len(fields) > 1:
                for i, f in enumerate(fields):
                    func = getattr(namespace.get("Collection"), f)
                    if func:
                        if not callable(func):
                            raise ValueError(f"{func} is not callable")
                        fields[i] = func

                    else:
                        errors.setdefault(coll_or_forward_ref, []).append(f)
        if errors:
            raise AttributeError(f"you must define the following Collection functions for distinction {dict(errors)}")

    @staticmethod
    def _build_model(relationships: Relationships, name: str):
        __edge_namespace__: dict[str, Any] = {}
        for field, relation_info in relationships.items():
            via_model = relation_info.via_model
            if relation_info.link_type in LIST_TYPES:
                if relation_info.link_type in (LinkTypes.OPTIONAL_EDGE_LIST, LinkTypes.OPTIONAL_LIST):
                    __edge_namespace__[field] = (Optional[list[via_model]], None)  # type: ignore[valid-type]
                else:
                    __edge_namespace__[field] = (list[via_model], ...)  # type: ignore[valid-type]

            elif relation_info.link_type in (LinkTypes.OPTIONAL_EDGE, LinkTypes.OPTIONAL_DIRECT):
                __edge_namespace__[field] = (Optional[via_model], None)
            else:
                __edge_namespace__[field] = (via_model, ...)  # type: ignore[assignment]
        m = create_model(f"{name}Edges", **__edge_namespace__, __base__=EdgeData)
        return m


class VertexModel(BaseArangoModel, Generic[TEdges], metaclass=VertexMeta):
    if TYPE_CHECKING:
        edges: TEdges
        __edge_to_field_mapping__: dict[Union[str, ForwardRef], list[str]] = {}

    class Collection(VertexCollectionConfig): ...

    def __init__(self, **data: Any):
        if EDGES in data:
            convert_edge_data_to_valid_kwargs(data[EDGES])

        super().__init__(**data)

        if EDGES not in data:  # note: enables dot notation for edges field
            object.__setattr__(self, EDGES, EdgeDict())

    def dict(
        self,
        *,
        include: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        exclude: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_edges: bool = False,
    ) -> "DictStrAny":
        if include_edges and self.__exclude_fields__:
            cast(dict, self.__exclude_fields__).pop("edges")

        try:
            super__dict = super().dict(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            )
        except RecursionError as e:
            raise AssertionError(
                "is not possible to call .dict() when using recursive model, instead traverse the graph and collect"
                " data or exclude recursive fields"
            ) from e

        if (
            self.__exclude_fields__ is None
            or EDGES not in self.__exclude_fields__
            and isinstance(self.edges, EdgeDict)
            and self.edges.__class__ == EdgeDict
            and not self.edges
        ):
            super__dict[EDGES] = None

        return super__dict

    def save_dict(self) -> "DictStrAny":
        return jsonable_encoder(self, by_alias=True, exclude=cast(set, self.__relationships_fields__.keys()))

    @classmethod
    def update_forward_refs(cls, **localns: Any) -> None:
        super().update_forward_refs(**localns)

        for k in cls.__edge_to_field_mapping__.copy():
            if isinstance(k, ForwardRef):
                funcs = cls.__edge_to_field_mapping__.pop(k)
                new_k = evaluate_forward_ref(cls, k, **localns)
                if new_k in cls.__edge_to_field_mapping__:
                    cls.__edge_to_field_mapping__[new_k.Collection.name].extend(funcs)
                else:
                    cls.__edge_to_field_mapping__[new_k.Collection.name] = funcs

        globalns = get_globals(cls)

        for fields, model_field in cls.__fields__[EDGES].type_.__fields__.items():
            if isinstance(model_field.type_, ForwardRef):
                model_field.type_ = evaluate_forwardref(model_field.type_, globalns, localns)

        cls.__fields__[EDGES].type_.update_forward_refs(**localns, **globalns)



import sys
from functools import lru_cache

from pydantic.typing import evaluate_forwardref


def get_globals(cls):
    if cls.__module__ in sys.modules:
        globalns = sys.modules[cls.__module__].__dict__.copy()
    else:
        globalns = {}
    return globalns


@lru_cache
def evaluate_forward_ref(source, model, **localns):
    return evaluate_forwardref(model, get_globals(source), localns)



from typing import TYPE_CHECKING, Any, Type, Union, cast

from pydango.connection.graph_utils import (
    EdgesIdsMapping,
    ModelFieldMapping,
    VerticesIdsMapping,
    _build_graph,
)
from pydango.connection.types import CollectionUpsertOptions, UpdateStrategy
from pydango.orm import ORMQuery
from pydango.orm.models import BaseArangoModel
from pydango.orm.query import for_
from pydango.query.consts import FROM, KEY, TO
from pydango.query.expressions import (
    IteratorExpression,
    RangeExpression,
    VariableExpression,
)
from pydango.query.functions import Length, Merge, UnionArrays
from pydango.query.options import UpsertOptions
from pydango.query.utils import new

if TYPE_CHECKING:
    from pydango.orm import VertexModel
    from pydango.query import AQLQuery


def _make_upsert_query(
    filter_: Any,
    i: Any,
    model: Union[Type[BaseArangoModel], BaseArangoModel],
    query: "AQLQuery",
    strategy: UpdateStrategy,
    options: Union[UpsertOptions, None] = None,
):
    if strategy == strategy.UPDATE:
        query = query.upsert(filter_, i, model.Collection.name, update=i, options=options)
    elif strategy == strategy.REPLACE:
        query = query.upsert(filter_, i, model.Collection.name, replace=i, options=options)

    return query


def _get_upsert_filter(
    document: Union["BaseArangoModel", VariableExpression], model: Union[Type["BaseArangoModel"], None] = None
):
    if not isinstance(document, BaseArangoModel) and model is not None:
        indexes = model.Collection.indexes
    elif isinstance(document, BaseArangoModel):
        indexes = document.Collection.indexes
    else:
        indexes = tuple()

    # todo: check first by _key or _id

    filter_ = {}
    for model_index in indexes:
        if hasattr(model_index, "unique") and model_index.unique:
            filter_ = {index_field: getattr(document, index_field) for index_field in model_index.fields}

        if isinstance(model_index, dict) and model_index.get("unique"):
            filter_ = {j: getattr(document, j) for j in model_index.get("fields", [])}

    if not filter_:
        key = getattr(document, KEY)
        filter_ = {"_key": key} if key is not None else {}  # noqa: PyProtectedMember

    return filter_


def _build_upsert_query(
    i: IteratorExpression,
    strategy: UpdateStrategy,
    model: Type["BaseArangoModel"],
    docs: Union[VariableExpression, list[VariableExpression]],
    *,
    edge: bool = False,
):
    filter_ = _get_upsert_filter(i, model)
    query = for_(i, in_=docs)
    query = _make_upsert_query(filter_, i, model, query, strategy, None).return_(new(edge=edge))
    return query


def _build_vertex_query(v, vertices_docs, strategy: UpdateStrategy):
    i = IteratorExpression()
    from_var = VariableExpression(v.Collection.name)
    query = _build_upsert_query(i, strategy, v, vertices_docs)
    return from_var, query


def _build_graph_query(
    document: "VertexModel",
    strategy: UpdateStrategy = UpdateStrategy.UPDATE,
    collection_options: Union[CollectionUpsertOptions, None] = None,
) -> tuple[ModelFieldMapping, VerticesIdsMapping, EdgesIdsMapping, ORMQuery]:
    query = ORMQuery()
    _visited: set[int] = set()
    edge_collections, edge_vertex_index, vertex_collections, model_fields_mapping = _build_graph(document, _visited)
    vertex_let_queries: dict[Type["VertexModel"], VariableExpression] = {}
    vertices_ids: VerticesIdsMapping = {}
    edge_ids: EdgesIdsMapping = {}
    for v in vertex_collections:
        vertex_docs = list(vertex_collections[v].values())
        vertices_ids[v] = {id(doc): i for i, doc in enumerate(vertex_docs)}
        from_var, vertex_query = _build_vertex_query(v, vertex_docs, strategy)
        vertex_let_queries[v] = from_var

        query.let(from_var, vertex_query)

    edge_let_queries = {}

    for e, coll in edge_vertex_index.items():
        counter = 0
        edge_vars = []
        for j, (instance, mapping) in enumerate(list(coll.items())):
            iterator = IteratorExpression()
            edge_ids.setdefault(e, {}).setdefault(instance, {}).update(
                {id(doc): i + counter for i, doc in enumerate(edge_collections[e][instance])}
            )
            edge_var_name = f"{e.Collection.name}_{j + 1}"
            edge = VariableExpression(edge_var_name)
            query.let(edge, edge_collections[e][instance])
            from_model: Type["VertexModel"]
            to_model: Type["VertexModel"]
            for k, ((from_model, to_model), rels) in enumerate(mapping.items()):
                from_ = vertex_collections[from_model].keys().index(instance)
                new_rels = [vertex_collections[to_model].keys().index(x) for x in rels]
                from_var = vertex_let_queries[from_model]
                to_var = vertex_let_queries[to_model]
                ret = {FROM: from_var[from_]._id, TO: to_var[iterator]._id}

                edge_from_to = VariableExpression(edge_var_name + f"_{k}_from_to")
                query.let(edge_from_to, for_(iterator, new_rels).return_(ret))

                merger = IteratorExpression("merger")

                merged = VariableExpression(edge_var_name + f"_{k}_merged")
                query.let(
                    merged,
                    for_(merger, RangeExpression(0, Length(edge_from_to) - 1)).return_(
                        Merge(edge[merger], edge_from_to[merger])
                    ),
                )
                edge_vars.append(merged)
                counter += len(rels)

        edges: Union[VariableExpression, list[VariableExpression]]
        if len(edge_vars) > 1:
            edges = cast(list[VariableExpression], UnionArrays(*edge_vars))
        elif len(edge_vars) == 1:
            edges = edge_vars[0]
        else:
            continue

        edge_iter = IteratorExpression()
        edge_let_queries[e] = VariableExpression(edge_var_name + "_result")
        query.let(edge_let_queries[e], _build_upsert_query(edge_iter, strategy, e, edges, edge=True))

    return (
        model_fields_mapping,
        vertices_ids,
        edge_ids,
        query.return_(
            {
                "vertex": {k.Collection.name: v for k, v in vertex_let_queries.items()},
                "edges": {k.Collection.name: v for k, v in edge_let_queries.items()},
            }
        ),
    )



from collections import OrderedDict, defaultdict
from typing import (
    Any,
    DefaultDict,
    Iterator,
    Optional,
    Type,
    Union,
    cast,
    get_args,
    get_origin,
)

from indexed import IndexedOrderedDict

from pydango.connection.types import (
    EdgeCollectionsMapping,
    EdgesIdsMapping,
    EdgeVerticesIndexMapping,
    ModelFieldMapping,
    RelationGroup,
    VertexCollectionsMapping,
    VerticesIdsMapping,
)
from pydango.orm.consts import EDGES
from pydango.orm.models import EdgeModel, VertexModel
from pydango.orm.models.base import LIST_TYPES, ArangoModel, BaseArangoModel, LazyProxy
from pydango.orm.models.utils import convert_edge_data_to_valid_kwargs
from pydango.orm.models.vertex import TVertexModel
from pydango.query.consts import FROM, ID, KEY, REV, TO


def get_collection_from_document(obj: Union[str, dict, "ArangoModel"]) -> str:
    _obj = None
    if isinstance(obj, dict):
        _obj = obj.get(ID)
    elif isinstance(obj, BaseArangoModel):
        _obj = obj.id

    if not _obj or not isinstance(_obj, str):
        raise ValueError("cannot parse collection")

    return _obj.partition("/")[0]


def _group_by_relation(
    model: BaseArangoModel,
) -> Iterator[RelationGroup]:
    relationships = model.__relationships__
    for field, relation in relationships.items():
        if get_origin(relation.link_model) is Union:
            for model_option in get_args(relation.link_model):
                yield RelationGroup(model_option.Collection.name, field, model_option, relation.via_model)
        else:
            yield RelationGroup(relation.link_model.Collection.name, field, relation.link_model, relation.via_model)


def _set_edge_operational_fields(result, model_id, edges_ids, i):
    e_obj = result["edges"][i.Collection.name][edges_ids[i.__class__][model_id][id(i)]]
    i.id = e_obj[ID]
    i.key = e_obj[KEY]
    i.rev = e_obj[REV]
    i.from_ = e_obj[FROM]
    i.to = e_obj[TO]


def db_traverse(
    model: TVertexModel,
    visited: set,
    result,
    model_fields_mapping: ModelFieldMapping,
    vertices_ids: VerticesIdsMapping,
    edges_ids: EdgesIdsMapping,
):
    model_id = id(model)
    if model_id in visited:
        return

    if isinstance(model, VertexModel):
        visited.add(model_id)

    v_index = vertices_ids[model.__class__][model_id]
    v_obj = result["vertex"][model.Collection.name][v_index]
    model.id = v_obj[ID]
    model.key = v_obj[KEY]
    model.rev = v_obj[REV]

    relations = list(_group_by_relation(model))
    if not relations:
        return

    for relation_group in relations:
        relation_doc: Union[TVertexModel, None] = getattr(model, relation_group.field)
        if not relation_doc:
            continue

        if isinstance(relation_doc, LazyProxy):
            relation_doc = relation_doc.__instance__

        if model.edges:
            for edge_field, obj in model.edges.__dict__.items():
                if isinstance(obj, list):
                    for i in obj:
                        _set_edge_operational_fields(result, model_id, edges_ids, i)
                elif obj is not None:
                    _set_edge_operational_fields(result, model_id, edges_ids, obj)
            if isinstance(relation_doc, list):
                z = zip(relation_doc, getattr(model.edges, relation_group.field, []))
                for vertex_doc, edge_doc in z:
                    db_traverse(vertex_doc, visited, result, model_fields_mapping, vertices_ids, edges_ids)
            else:
                getattr(model.edges, relation_group.field)
                db_traverse(
                    cast(VertexModel, relation_doc), visited, result, model_fields_mapping, vertices_ids, edges_ids
                )
        else:
            # todo: insert join relation
            raise NotImplementedError("join relation not implemented yet")


def graph_to_document(traversal_result: dict, model: Type[VertexModel]):
    doc = traversal_result["doc"]

    vertices: DefaultDict[str, dict[str, Any]] = defaultdict(dict)
    edges: dict[str, dict[tuple[str, str], Union[list[dict[str, Any]], dict[str, Any]]]] = {}
    if doc:
        vertices[doc[ID]] = doc
    edge_count = 0
    for relation in traversal_result["edges"]:
        v = relation["v"]
        e = relation["e"]
        edge_coll = get_collection_from_document(e)
        vertices[v[ID]] = v
        coordinate = (e[FROM], e[TO])

        if coordinate not in edges:
            edges.setdefault(edge_coll, {})[coordinate] = e
        elif isinstance(edges[edge_coll][coordinate], list):
            cast(list, edges[edge_coll][coordinate]).append(e)
        else:
            edges.setdefault(edge_coll, {})[coordinate] = [cast(dict[str, Any], edges[edge_coll][coordinate])]
            cast(list, edges[edge_coll][coordinate]).append(e)

        edge_count += 1

    if len(traversal_result["edges"]) != edge_count:
        raise AssertionError("something happened could not map all edges")

    new_d, recursive = map_graph_edges({"start": doc[ID], "vertices": vertices, "edges": edges}, model)

    return new_d, recursive


def map_graph_edges(graph, model):
    vertices = graph["vertices"]
    edges = graph["edges"]
    start = graph["start"]
    visited = set()
    recursive = False
    for coll, _edges in edges.items():
        for (f, t), e in _edges.items():
            to = vertices[t]
            coll, _, __ = e[ID].partition("/")

            for func in model.__edge_to_field_mapping__[coll]:
                if to[ID] == start:
                    continue
                if callable(func):
                    if func(e, to):
                        map_edge(e, f, func.__name__, model, to, vertices)
                        break
                else:
                    map_edge(e, f, func, model, to, vertices)

                if id(to) in visited:
                    recursive = True
                visited.add(id(to))
    # note here to detect circular references
    #  - not supported in pydantic v1 start in pydantic v2
    return vertices[start], recursive


def map_edge(e, f, func, model, to, vertices):
    link_type = model.__relationships__[func].link_type
    if link_type in LIST_TYPES:
        if func not in vertices[f]:
            vertices[f][func] = []

        vertices[f][func].append(to)

        if func not in vertices[f].setdefault(EDGES, {}):
            vertices[f][EDGES][func] = []
        vertices[f][EDGES][func].append(e)
    else:
        vertices[f][func] = to
        vertices[f].setdefault(EDGES, {})[func] = e


def _build_object_graph_mappings(
    model: "TVertexModel",
    visited: set[int],
    vertex_collections: VertexCollectionsMapping,
    edge_collections: EdgeCollectionsMapping,
    edge_vertex_index: EdgeVerticesIndexMapping,
    model_fields_mapping: ModelFieldMapping,
) -> None:
    def _add_model_field_to_mapping(model, field, relation_doc, edge_doc):
        model_id = id(model)

        mapping = model_fields_mapping.setdefault(model.__class__, {})
        model_mapping = mapping.setdefault(model_id, {})

        if model.__relationships__[field].link_type in LIST_TYPES:
            model_mapping.setdefault(field, []).append({"v": id(relation_doc), "e": id(edge_doc)})
        else:
            model_mapping[field] = {"v": id(relation_doc), "e": id(edge_doc)}

    def _prepare_relation(field, model, edge_cls, edge_doc, relation_doc):
        model_id = id(model)
        if id(relation_doc) in (
            edge_vertex_index.setdefault(edge_cls, {})
            .setdefault(model_id, {})
            .setdefault((model.__class__, relation_doc.__class__), [])
        ):
            return False

        if edge_doc:
            edge_collections.setdefault(edge_cls, IndexedOrderedDict()).setdefault(model_id, []).append(edge_doc)

        _add_model_field_to_mapping(model, field, relation_doc, edge_doc)

        (
            edge_vertex_index.setdefault(edge_cls, {})
            .setdefault(model_id, {})
            .setdefault((model.__class__, relation_doc.__class__), [])
            .append(id(relation_doc))
        )

    if id(model) in visited:
        return

    if isinstance(model, VertexModel):
        vertex_collections.setdefault(model.__class__, IndexedOrderedDict())[id(model)] = model
        visited.add(id(model))

    relations = list(_group_by_relation(model))

    if not relations:
        return

    for relation_group in relations:
        relation_doc: Union[TVertexModel, None] = getattr(model, relation_group.field)

        if isinstance(relation_doc, LazyProxy):
            relation_doc = cast(VertexModel, relation_doc.__instance__)

        if not relation_doc:
            _add_model_field_to_mapping(model, relation_group.field, None, None)
            continue

        edge_cls: Optional[Type[EdgeModel]] = relation_group.via_model

        if model.edges:
            if isinstance(model.edges, dict):
                convert_edge_data_to_valid_kwargs(model.edges)
                # todo: this initiate the class edge model so it validates the edges, should we do that?
                model.edges = model.__fields__[EDGES].type_(**model.edges)

            if isinstance(relation_doc, list):
                if len(getattr(model.edges, relation_group.field, [])) != len(relation_doc):
                    raise AssertionError(
                        f"{model.__class__.__name__} vertex edges {relation_group.field} number mismatch"
                    )
                vertex_doc: VertexModel
                edge_doc: EdgeModel
                for vertex_doc, edge_doc in zip(relation_doc, getattr(model.edges, relation_group.field, [])):
                    _prepare_relation(relation_group.field, model, edge_cls, edge_doc, vertex_doc)
                    _build_object_graph_mappings(
                        vertex_doc,
                        visited,
                        vertex_collections,
                        edge_collections,
                        edge_vertex_index,
                        model_fields_mapping,
                    )

            else:
                edge_doc = getattr(model.edges, relation_group.field)
                _prepare_relation(relation_group.field, model, edge_cls, edge_doc, relation_doc)
                _build_object_graph_mappings(
                    relation_doc, visited, vertex_collections, edge_collections, edge_vertex_index, model_fields_mapping
                )
        else:
            # todo: insert join relation
            pass


def _build_graph(
    document: VertexModel, _visited: set[int]
) -> tuple[EdgeCollectionsMapping, EdgeVerticesIndexMapping, VertexCollectionsMapping, ModelFieldMapping]:
    vertex_collections: VertexCollectionsMapping = OrderedDict()
    edge_collections: EdgeCollectionsMapping = OrderedDict()
    edge_vertex_index: EdgeVerticesIndexMapping = {}  # defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    model_fields_mapping: ModelFieldMapping = {}

    _build_object_graph_mappings(
        document, _visited, vertex_collections, edge_collections, edge_vertex_index, model_fields_mapping
    )
    return edge_collections, edge_vertex_index, vertex_collections, model_fields_mapping



from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union, cast

from pydantic.fields import ModelField

from pydango.orm.models.sentinel import LazyFetch
from pydango.query.expressions import (
    Expression,
    FieldExpression,
    IteratorExpression,
    VariableExpression,
)

if TYPE_CHECKING:
    from pydantic.fields import LocStr, ModelOrDc, ValidateReturn

    from pydango.orm.models.vertex import TVertexModel
    from pydango.query.expressions import QueryExpression


class ModelFieldExpression(FieldExpression):
    def __init__(self, field: Union[str, Expression], parent: Type["TVertexModel"]):
        super().__init__(field, cast(VariableExpression, parent))
        self.parent = parent  # type: ignore[assignment]

    def compile(self, query_ref: "QueryExpression") -> str:
        if isinstance(self.field, Expression):
            return super().compile(query_ref)
        else:
            if not isinstance(self.parent, IteratorExpression):
                # currently importing ORMQuery creates a circular dependency
                compiled = query_ref.orm_bound_vars[self.parent]  # type: ignore[attr-defined]
                return f"{compiled.compile(query_ref)}.{self.field}"
            return super().compile(query_ref)

    def __hash__(self):
        return hash(self.field)


class RelationModelField(ModelField):
    def validate(
        self,
        v: Any,
        values: Dict[str, Any],
        *,
        loc: "LocStr",
        cls: Optional["ModelOrDc"] = None,
    ) -> "ValidateReturn":
        return super().validate(v, values, loc=loc, cls=cls) if not isinstance(v, LazyFetch) else (v, None)


def get_pydango_field(field: ModelField, cls: Type[RelationModelField] = RelationModelField) -> RelationModelField:
    return cls(
        name=field.name,
        type_=field.annotation,
        alias=field.alias,
        class_validators=field.class_validators,
        default=field.default,
        default_factory=field.default_factory,
        required=field.required,
        model_config=field.model_config,
        final=field.final,
        field_info=field.field_info,
    )



from typing import TYPE_CHECKING, ForwardRef, Optional, Type

if TYPE_CHECKING:
    from pydantic.fields import ModelField
    from pydantic.typing import ReprArgs

    from pydango.orm.models.base import LinkTypes
    from pydango.orm.models.edge import TEdge
    from pydango.orm.models.vertex import TVertexModel


class Relationship:
    def __init__(
        self,
        *,
        field: "ModelField",
        back_populates: Optional[str] = None,
        link_model: Type["TVertexModel"],
        via_model: Optional[Type["TEdge"]] = None,
        link_type: "LinkTypes",
    ):
        self.via_model = via_model
        self.link_type = link_type
        self.field = field
        self.link_model = link_model
        self.back_populates = back_populates

    def __repr_args__(self) -> "ReprArgs":
        name = self.link_model.__name__ if not isinstance(self.link_model, ForwardRef) else self.link_model
        args = [("link_model", name), ("link_type", self.link_type.value)]
        if self.via_model:
            args.append(("via_model", self.via_model.__name__))
        return args



from typing import Final

KEY: Final[str] = "_key"
"""The key identifier used in the query."""

ID: Final[str] = "_id"
"""The ID identifier used in the query."""

REV: Final[str] = "_rev"
"""The revision identifier used in the query."""

FROM: Final[str] = "_from"
"""The 'from' field used in query relations."""

TO: Final[str] = "_to"
"""The 'to' field used in query relations."""

DYNAMIC_ALIAS: Final[str] = "DynamicAlias"
"""Alias used for dynamic query operations."""



from pydantic.fields import (
    SHAPE_FROZENSET,
    SHAPE_ITERABLE,
    SHAPE_LIST,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_TUPLE_ELLIPSIS,
)

LIST_SHAPES = {
    SHAPE_LIST,
    SHAPE_TUPLE_ELLIPSIS,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_FROZENSET,
    SHAPE_ITERABLE,
}



import sys
from typing import ForwardRef, Union

from pydango.orm.models.fields import RelationModelField
from pydango.orm.models.relations import Relationship

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

RelationshipFields: TypeAlias = dict[str, RelationModelField]
Relationships: TypeAlias = dict[str, Relationship]
EdgeFieldMapping: TypeAlias = dict[Union[str, ForwardRef], list[str]]



import sys
from abc import ABC
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
    overload,
)

from pydango.query.consts import KEY
from pydango.query.expressions import (
    AssignmentExpression,
    CollectionExpression,
    FieldExpression,
    IterableExpression,
    IteratorExpression,
    ListExpression,
    LiteralExpression,
    ObjectExpression,
    QueryExpression,
    RangeExpression,
    SortDirection,
    SortExpression,
    VariableExpression,
)
from pydango.query.utils import Compilable

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from pydango.query.expressions import (
        Expression,
        LogicalExpression,
        ReturnableExpression,
    )
    from pydango.query.options import (
        BaseModificationOptions,
        CollectOptions,
        LoopOptions,
        RemoveOptions,
        ReplaceOptions,
        UpdateOptions,
        UpsertOptions,
    )
    from pydango.query.query import AQLQuery


class Operation(Compilable):
    def __init__(self, query_ref: "QueryExpression"):
        self.query_ref = query_ref

    def compile(self, *args, **kwargs):
        raise NotImplementedError


ForParams = Union[str, IteratorExpression, IterableExpression, "AQLQuery"]

MANUAL_TYPES = (str, IteratorExpression)
AUTOMATIC_TYPES = (
    ListExpression,
    CollectionExpression,
    # Aliased,
)


class ForOperation(Operation):
    def __init__(
        self,
        collection_or_variable: ForParams,
        in_: Optional[Union[list, IterableExpression]] = None,
        *,
        query_ref: "QueryExpression",
        options: Optional["LoopOptions"] = None,
    ):
        super().__init__(query_ref)
        self.options = options
        self.query_ref = query_ref
        self.variable: VariableExpression
        if isinstance(collection_or_variable, MANUAL_TYPES):
            if isinstance(in_, list):
                in_ = ListExpression(tuple(in_))
            elif isinstance(in_, tuple):
                in_ = ListExpression(in_)

            elif in_ is None:
                raise AssertionError(f"you must provide in_ field when using {MANUAL_TYPES}")
            elif not isinstance(in_, (IterableExpression, VariableExpression)):
                raise AssertionError(f"in_ should be instance of {IterableExpression.__name__}")

            if isinstance(collection_or_variable, IteratorExpression):
                if collection_or_variable.var_name in query_ref.__used_vars__:
                    raise AttributeError(f"{collection_or_variable.var_name} variable is already defined")

                if collection_or_variable.var_name:
                    self.query_ref.__used_vars__.add(collection_or_variable.var_name)

            if isinstance(collection_or_variable, str):
                collection_or_variable = IteratorExpression(collection_or_variable)

            self.variable = collection_or_variable
            self.in_ = in_

        elif isinstance(collection_or_variable, AUTOMATIC_TYPES):
            if isinstance(collection_or_variable, CollectionExpression):
                if not collection_or_variable.iterator:
                    i = IteratorExpression()
                    collection_or_variable.iterator = i

                elif collection_or_variable.iterator.var_name in query_ref.__used_vars__:
                    raise ValueError(f"{collection_or_variable.iterator.var_name} variable is already defined")
                elif collection_or_variable.iterator.var_name:
                    query_ref.__used_vars__.add(collection_or_variable.iterator.var_name)

                self.in_ = cast(IterableExpression, collection_or_variable)
                self.variable = (
                    collection_or_variable.iterator
                )  # or IteratorExpression(collection_or_variable.value) todo: check this case

            elif isinstance(collection_or_variable, ListExpression):
                if not collection_or_variable.iterator:
                    i = IteratorExpression()
                    collection_or_variable.iterator = i
                elif collection_or_variable.iterator.var_name in query_ref.__used_vars__:
                    raise ValueError(f"{collection_or_variable.iterator.var_name} variable is already defined")
                elif collection_or_variable.iterator.var_name:
                    query_ref.__used_vars__.add(collection_or_variable.iterator.var_name)

                self.in_ = collection_or_variable
                self.variable = collection_or_variable.iterator or IteratorExpression(collection_or_variable.value)
        # todo: check this case
        # else:
        #     self.variable = collection_or_variable
        #     self.in_ = in_

    def compile(self) -> str:
        compiled = f"FOR {self.variable.compile(self.query_ref)} IN {self.in_.compile(self.query_ref)}"
        if self.options:
            options_compile = self.options.compile()
            compiled += options_compile and f"OPTIONS {options_compile}"
        return compiled

    def __repr__(self):
        _repr = f"FOR {self.variable} IN {self.in_}"
        if self.options:
            options_compile = self.options.compile()
            _repr += options_compile and f"OPTIONS {options_compile}"
        return _repr


class TraversalDirection(str, Enum):
    OUTBOUND = "OUTBOUND"
    INBOUND = "INBOUND"
    ANY = "ANY"


class TraversalOperation(Operation):
    def __init__(
        self,
        iterators: Union[
            IteratorExpression,
            list[IteratorExpression],
            tuple[IteratorExpression],
            tuple[IteratorExpression, IteratorExpression],
            tuple[IteratorExpression, IteratorExpression, IteratorExpression],
            tuple[IteratorExpression, ...],
        ],
        edges: Union[str, CollectionExpression, Sequence[Union[str, CollectionExpression]]],
        start: Union[str, "LiteralExpression", VariableExpression, FieldExpression],
        depth: Union[range, tuple[int, int], "RangeExpression"],
        direction: TraversalDirection,
        query_ref: "AQLQuery",
    ):
        super().__init__(query_ref)
        if isinstance(edges, str):
            edges = CollectionExpression(edges)
        elif isinstance(edges, Sequence):
            edges = [CollectionExpression(i) if isinstance(i, str) else i for i in edges]
            # for i in enumerate(edges):
            #     if isinstance(i,str):

        if isinstance(iterators, IteratorExpression):
            iterators = (iterators,)
        if isinstance(iterators, list):
            iterators = tuple(iterators)

        # handle depth
        if isinstance(depth, tuple):
            depth = RangeExpression(depth[0], depth[1])
        elif isinstance(depth, range):
            depth = RangeExpression(depth.start, depth.stop)

        if isinstance(start, str):
            start = LiteralExpression(start)
        self.iterators = iterators
        self.edges = edges
        self.start = start
        self.direction = direction
        self.depth = depth

    def compile(self, *args, **kwargs):
        compiled_iterators = []
        for i in self.iterators:
            compiled_iterators.append(i.compile(self.query_ref))
        edges = (
            isinstance(self.edges, list)
            and ", ".join([i.compile(self.query_ref) for i in self.edges])
            or self.edges.compile(self.query_ref)
        )
        return (
            f"FOR {', '.join(compiled_iterators)} IN "
            f"{self.depth.compile(self.query_ref)} {self.direction.value} {self.start.compile(self.query_ref)} "
            f"{edges}"
        )

    def __repr__(self):
        compiled_iterators = []
        for i in self.iterators:
            compiled_iterators.append(repr(i))

        return (
            f"FOR {', '.join(compiled_iterators)} IN"
            f" {repr(self.depth)} {self.direction.value} {self.start} {self.edges}"
        )


class LetOperation(Operation):
    def __init__(
        self,
        variable: Union[str, VariableExpression],
        expression: "Expression",
        *,
        query_ref: "AQLQuery",
    ):
        super().__init__(query_ref=query_ref)
        if isinstance(variable, str):
            variable = VariableExpression(variable)

        if variable.var_name:
            # self.query_ref.__dynamic_vars__.append(variable)
            # else:
            self.query_ref.__used_vars__.add(variable.var_name)

        if isinstance(expression, QueryExpression):
            expression.parent = self.query_ref
            self._is_query = True

        if isinstance(expression, list):
            expression = ListExpression(expression)

        if isinstance(expression, dict):
            expression = ObjectExpression(expression)

        self.expression = AssignmentExpression(variable, expression)

    def compile(self, *args, **kwargs):
        if isinstance(self.expression, AssignmentExpression) and isinstance(
            self.expression.expression, QueryExpression
        ):
            self.expression.expression.sep = " "
        return f"LET {self.expression.compile(self.query_ref)}"

    def __repr__(self):
        return f"LET {repr(self.expression)}"


class FilterOperation(Operation):
    def __init__(self, condition: "LogicalExpression", query_ref: "AQLQuery"):
        super().__init__(query_ref=query_ref)
        self.condition = condition

    def compile(self):
        return f"FILTER {self.condition.compile(self.query_ref)}"

    def __repr__(self):
        return f"FILTER {repr(self.condition)}"


SortParams = Union[SortExpression, FieldExpression, tuple[FieldExpression, SortDirection]]


class SortOperation(Operation):
    def __init__(self, *sort_list: SortParams, query_ref: "AQLQuery"):
        super().__init__(query_ref=query_ref)
        self.sort_list = sort_list

    def compile(self, *args, **kwargs) -> str:
        comp = []
        for expr in self.sort_list:
            if isinstance(expr, tuple):
                comp.append(SortExpression(expr[0], expr[1]).compile(self.query_ref))
            elif isinstance(expr, FieldExpression):
                comp.append(SortExpression(expr).compile(self.query_ref))
            else:
                comp.append(expr.compile(self.query_ref))
        return f"SORT {', '.join(comp)}"

    def __repr__(self):
        return f"SORT {', '.join(repr(i) for i in self.sort_list)}"


class ReturnOperation(Operation):
    def __init__(self, return_expr: Union[dict, "ReturnableExpression"], query_ref: "AQLQuery", *, distinct=None):
        super().__init__(query_ref=query_ref)

        if isinstance(return_expr, CollectionExpression):
            return_expr = return_expr.iterator

        elif isinstance(return_expr, list):
            return_expr = ListExpression(return_expr)
            # for i in return_expr:
            #     if isinstance(i,str)
            #     if not isinstance(i, FieldExpression):
            #         raise Exception("todo: check this")
        elif isinstance(return_expr, Mapping):
            return_expr = ObjectExpression(return_expr)

        self.distinct = distinct
        self.return_expr = return_expr
        self.query_ref = query_ref

    def compile(self, *args, **kwargs):
        start = "RETURN"
        if self.distinct:
            start += " DISTINCT"

        return f"{start} {self.return_expr.compile(self.query_ref)}"

    def __repr__(self):
        start = "RETURN"
        if self.distinct:
            start += " DISTINCT"
        return f"{start} {repr(self.return_expr)}"


class LimitOperation(Operation):
    def __init__(
        self,
        limit: int,
        query_ref: "AQLQuery",
        offset=None,
    ):
        super().__init__(query_ref=query_ref)
        self.offset = offset
        self.limit = limit

    def compile(self, *args, **kwargs):
        if self.offset:
            return f"LIMIT {self.offset},{self.limit}"
        else:
            return f"LIMIT {self.limit}"

    def __repr__(self):
        return f"LIMIT {self.limit}"


class InsertOperation(Operation):
    def __init__(
        self,
        doc: Union[dict, ObjectExpression, VariableExpression],
        collection: Union[str, CollectionExpression],
        query_ref: "AQLQuery",
    ):
        super().__init__(query_ref=query_ref)
        if isinstance(doc, dict):
            doc = ObjectExpression(doc)
        if isinstance(collection, str):
            collection = CollectionExpression(collection)
        self.collection = collection
        self.doc = doc

    def compile(self, *args, **kwargs):
        return f"INSERT {self.doc.compile(self.query_ref)} INTO {self.collection.compile(self.query_ref)}"

    def __repr__(self):
        return f"INSERT {repr(self.doc)} INTO {repr(self.collection)}"


class RemoveOperation(Operation):
    def __init__(
        self,
        expression: Union[str, dict, LiteralExpression, FieldExpression, VariableExpression, ObjectExpression],
        collection: Union[str, CollectionExpression],
        *,
        options: Optional["RemoveOptions"] = None,
        query_ref: "AQLQuery",
    ):
        super().__init__(query_ref=query_ref)
        self.options = options

        if isinstance(expression, dict):
            if KEY not in expression:
                raise ValueError(f"expression must contain {KEY}")

            expression = ObjectExpression({KEY: expression[KEY]})
        if isinstance(expression, str):
            expression = LiteralExpression(expression)
        if isinstance(collection, str):
            collection = CollectionExpression(collection)
        self.expression = expression
        self.collection = collection

    def compile(self, *args, **kwargs):
        compiled = f"REMOVE {self.expression.compile(self.query_ref)} IN {self.collection.compile(self.query_ref)}"
        if self.options:
            options_compile = self.options.compile()
            compiled += options_compile and f"OPTIONS {options_compile}"
        return compiled

    def __repr__(self):
        _repr = f"REMOVE {repr(self.expression)} IN {repr(self.collection)}"
        if self.options:
            options_compile = self.options.compile()
            _repr += options_compile and f"OPTIONS {options_compile}"
        return _repr


class BaseChangeOperation(Operation, ABC):
    _keyword: str

    def __init__(
        self,
        key: Union[str, dict, ObjectExpression],
        obj: Union[ObjectExpression, dict],
        collection: Union[CollectionExpression, str],
        *,
        options: Optional["BaseModificationOptions"],
        query_ref,
    ):
        super().__init__(query_ref=query_ref)
        if isinstance(key, dict):
            key = ObjectExpression(key)
        if isinstance(obj, dict):
            obj = ObjectExpression(obj)
        if isinstance(collection, str):
            collection = CollectionExpression(collection)
        self.options = options
        self.key = key
        self.obj = obj
        self.collection = collection

    def compile(self, *args, **kwargs):
        compiled = f"{self._keyword} {self.obj.compile(self.query_ref)} IN {self.collection.compile(self.query_ref)}"
        if self.options:
            options_compile = self.options.compile()
            compiled += options_compile and f"OPTIONS {options_compile}"
        return compiled

    def __repr__(self):
        _repr = f"{self._keyword} {repr(self.obj)} IN {repr(self.collection)}"
        if self.options:
            options_compile = self.options.compile()
            _repr += options_compile and f"OPTIONS {options_compile}"
        return _repr


class UpdateOperation(BaseChangeOperation):
    _keyword = "UPDATE"

    def __init__(
        self,
        key: Union[str, dict, ObjectExpression],
        obj: Union[ObjectExpression, dict],
        collection: Union[CollectionExpression, str],
        *,
        options: Optional["UpdateOptions"],
        query_ref,
    ):
        super().__init__(key, obj, collection, options=options, query_ref=query_ref)


class ReplaceOperation(BaseChangeOperation):
    _keyword = "REPLACE"

    def __init__(
        self,
        key: Union[str, dict, ObjectExpression],
        obj: Union[ObjectExpression, dict],
        collection: Union[CollectionExpression, str],
        *,
        options: Optional["ReplaceOptions"],
        query_ref,
    ):
        super().__init__(key, obj, collection, options=options, query_ref=query_ref)


class UpsertOperation(Operation):
    @overload
    def __init__(
        self,
        query_ref: "AQLQuery",
        filter_: Union[dict, ObjectExpression, VariableExpression],
        collection: Union[str, CollectionExpression],
        insert: Union[dict, ObjectExpression, VariableExpression],
        *,
        update: Union[dict, ObjectExpression, VariableExpression],
        options: Optional["UpsertOptions"] = None,
    ): ...

    @overload
    def __init__(
        self,
        query_ref: "AQLQuery",
        filter_: Union[dict, ObjectExpression, VariableExpression],
        collection: Union[str, CollectionExpression],
        insert: Union[dict, ObjectExpression, VariableExpression],
        *,
        replace: Union[dict, ObjectExpression, VariableExpression],
        options: Optional["UpsertOptions"] = None,
    ): ...

    def __init__(
        self,
        query_ref: "AQLQuery",
        filter_: Union[dict, ObjectExpression, VariableExpression],
        collection: Union[str, CollectionExpression],
        insert: Union[dict, ObjectExpression, VariableExpression],
        *,
        update: Union[dict, ObjectExpression, VariableExpression, None] = None,
        replace: Union[dict, ObjectExpression, VariableExpression, None] = None,
        options: Optional["UpsertOptions"] = None,
    ):
        super().__init__(query_ref)
        self.options = options
        if update and replace:
            raise ValueError("cannot set both update and replace")
        if not (update or replace):
            raise ValueError("you need to specify update or replace")
        if isinstance(replace, dict):
            replace = ObjectExpression(replace)

        if isinstance(update, dict):
            update = ObjectExpression(update)

        if isinstance(insert, dict):
            insert = ObjectExpression(insert)
        if isinstance(collection, str):
            collection = CollectionExpression(collection)

        if isinstance(filter_, dict):
            filter_ = ObjectExpression(filter_)

        self.replace = replace
        self.filter = filter_
        self.collection = collection
        self.update = update
        self.insert = insert

    def compile(self, *args, **kwargs):
        if self.update:
            modification = f"UPDATE {self.update.compile(self.query_ref)}"
        else:
            modification = f"REPLACE {self.replace.compile(self.query_ref)}"
        compiled = (
            f"UPSERT {self.filter.compile(self.query_ref)} INSERT {self.insert.compile(self.query_ref)} "
            f"{modification} IN {self.collection.compile(self.query_ref)}"
        )
        if self.options:
            options_compile = self.options.compile()
            compiled += options_compile and f"OPTIONS {options_compile}"
        return compiled

    def __repr__(self):
        if self.update:
            modification = f"UPDATE {repr(self.update)}"
        else:
            modification = f"REPLACE {repr(self.replace)}"
        _repr = f"UPSERT {repr(self.filter)} INSERT {repr(self.insert)} {modification} IN {repr(self.collection)}"

        if self.options:
            options_compile = self.options.compile()
            _repr += options_compile and f"OPTIONS {options_compile}"
        return _repr


AssignmentParam: TypeAlias = Union[AssignmentExpression, tuple[VariableExpression, "Expression"]]
AssignmentParams: TypeAlias = Union[AssignmentParam, Sequence[AssignmentParam]]
IntoParam = Union[
    VariableExpression,
    AssignmentExpression,
]


class CollectOperation(Operation):
    def __init__(
        self,
        *,
        collect: Optional[AssignmentParams] = None,
        aggregate: Optional[AssignmentParams] = None,
        into: Union[VariableExpression, AssignmentParam, None] = None,
        keep: Optional[VariableExpression] = None,
        with_count_into: Optional[VariableExpression] = None,
        options: Optional["CollectOptions"] = None,
        query_ref: "AQLQuery",
    ):
        if not any((collect, aggregate, with_count_into)):
            raise ValueError("you should use one of collect, aggregate or count")
        if aggregate:
            if with_count_into:
                raise ValueError("aggregate cannot be used in conjunction of WITH COUNT IT")
            if keep:
                raise ValueError("cannot use KEEP with aggregate")

        super().__init__(query_ref)
        if collect and not isinstance(collect, Sequence):
            collect = [collect]
        if aggregate and not isinstance(aggregate, Sequence):
            aggregate = [aggregate]

        params: list[Any] = [collect or [], aggregate or []]
        for param in params:
            for i, assignment in enumerate(param):
                if isinstance(assignment, tuple):
                    param[i] = AssignmentExpression(assignment[0], assignment[1])
                if assignment.variable.var_name:
                    self.query_ref.__used_vars__.add(assignment.variable.var_name)

        for param in [into, with_count_into]:
            if isinstance(param, VariableExpression):
                if param.var_name:
                    self.query_ref.__used_vars__.add(param.var_name)
            elif isinstance(param, AssignmentExpression):
                if param.variable.var_name:
                    self.query_ref.__used_vars__.add(param.variable.var_name)

                if isinstance(param.expression, ObjectExpression):
                    # TODO: handle object
                    pass

                elif isinstance(param.expression, ListExpression):
                    # TODO: handle list
                    pass

        self.collect = collect
        self.aggregate = aggregate
        self.into = into
        self.keep = keep
        self.with_count_into = with_count_into
        self.options = options

        params = [collect, aggregate, into, keep, with_count_into]
        keywords = ["COLLECT", "AGGREGATE", "INTO", "KEEP", "WITH COUNT INTO"]
        self.to_compile = [i for i in list(zip(keywords, params))[2:] if i[1] is not None]

    def compile(self):
        compiled = self.collect and f"COLLECT {', '.join(i.compile(self.query_ref) for i in self.collect)}" or "COLLECT"
        agg = self.aggregate and f"AGGREGATE {', '.join(i.compile(self.query_ref) for i in self.aggregate)}" or None
        compiled = agg and " ".join((compiled, agg)) or compiled

        to_compile = " ".join([f"{k} {v.compile(self.query_ref)}" for k, v in self.to_compile])
        return f"{compiled}" + (to_compile and f" {to_compile}" or "")

    def __repr__(self):
        compiled = self.collect and f"COLLECT {', '.join(repr(i) for i in self.collect)}" or "COLLECT"
        agg = self.aggregate and f"AGGREGATE {', '.join(repr(i) for i in self.aggregate)}" or None
        compiled = agg and " ".join((compiled, agg)) or compiled

        to_compile = " ".join([f"{k} {repr(v)}" for k, v in self.to_compile])
        return f"{compiled}" + (to_compile and f" {to_compile}" or "")


class WindowOperation(Operation):
    def compile(self, *args, **kwargs):
        # TODO
        pass


class WithOperation(Operation):
    def compile(self, *args, **kwargs):
        # TODO
        pass



import json
from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from pydango.query.utils import Compilable


class Options(Compilable):
    _map: dict = {}

    def compile(self):
        pairs = []

        for field, value in self._map.items():
            if value is None:
                continue
            pairs.append(f"{field}: {json.dumps(value)}")
        if pairs:
            return f"{{{', '.join(pairs)}}}"


@dataclass
class LoopOptions(Options):
    index_hint: Optional[Union[list[str], str]] = None
    force_index_hint: Optional[bool] = None
    disable_index: Optional[bool] = None
    max_projections: Optional[int] = None
    use_cache: Optional[bool] = None
    look_ahead: Optional[int] = None

    def __post_init__(self):
        self._map = {
            "IndexHint": self.index_hint,
            "forceIndexHint": self.force_index_hint,
            "disableIndex": self.disable_index,
            "maxProjections": self.max_projections,
            "useCache": self.use_cache,
            "lookAhead": self.look_ahead,
        }


# noinspection DuplicatedCode
@dataclass
class BaseModificationOptions(Options, ABC):
    ignore_errors: Optional[bool] = None
    wait_for_sync: Optional[bool] = None
    ignore_revs: Optional[bool] = None
    exclusive: Optional[bool] = None
    refill_index_caches: Optional[bool] = None

    def __post_init__(self):
        self._map = {
            "ignoreErrors": self.ignore_errors,
            "waitForSync": self.wait_for_sync,
            "ignoreRevs": self.ignore_revs,
            "exclusive": self.exclusive,
            "refillIndexCaches": self.refill_index_caches,
        }


# noinspection DuplicatedCode
@dataclass()
class RemoveOptions(BaseModificationOptions):
    ignore_errors: Optional[bool] = None
    wait_for_sync: Optional[bool] = None
    ignore_revs: Optional[bool] = None
    exclusive: Optional[bool] = None
    refill_index_caches: Optional[bool] = None

    def __post_init__(self):
        self._map = {
            "ignoreErrors": self.ignore_errors,
            "waitForSync": self.wait_for_sync,
            "ignoreRevs": self.ignore_revs,
            "exclusive": self.exclusive,
            "refillIndexCaches": self.refill_index_caches,
        }


@dataclass()
class UpdateOptions(BaseModificationOptions):
    ignore_errors: Optional[bool] = None
    keep_null: Optional[bool] = None
    merge_objects: Optional[bool] = None
    wait_for_sync: Optional[bool] = None
    ignore_revs: Optional[bool] = None
    exclusive: Optional[bool] = None
    refill_index_caches: Optional[bool] = None

    def __post_init__(self):
        self._map = {
            "ignoreErrors": self.ignore_errors,
            "keepNull": self.keep_null,
            "mergeObjects": self.merge_objects,
            "waitForSync": self.wait_for_sync,
            "ignoreRevs": self.ignore_revs,
            "exclusive": self.exclusive,
            "refillIndexCaches": self.refill_index_caches,
        }


# noinspection DuplicatedCode
@dataclass()
class ReplaceOptions(BaseModificationOptions):
    ignore_errors: Optional[bool] = None
    wait_for_sync: Optional[bool] = None
    ignore_revs: Optional[bool] = None
    exclusive: Optional[bool] = None
    refill_index_caches: Optional[bool] = None

    def __post_init__(self):
        self._map = {
            "ignoreErrors": self.ignore_errors,
            "waitForSync": self.wait_for_sync,
            "ignoreRevs": self.ignore_revs,
            "exclusive": self.exclusive,
            "refillIndexCaches": self.refill_index_caches,
        }


@dataclass()
class UpsertOptions(BaseModificationOptions):
    ignore_errors: Optional[bool] = None
    keep_null: Optional[bool] = None
    merge_objects: Optional[bool] = None
    wait_for_sync: Optional[bool] = None
    ignore_revs: Optional[bool] = None
    exclusive: Optional[bool] = None
    index_hint: Optional[Union[list[str], str]] = None
    force_index_hint: Optional[bool] = None

    def __post_init__(self):
        self._map = {
            "ignoreErrors": self.ignore_errors,
            "keepNull": self.keep_null,
            "mergeObjects": self.merge_objects,
            "waitForSync": self.wait_for_sync,
            "ignoreRevs": self.ignore_revs,
            "exclusive": self.exclusive,
            "indexHint": self.index_hint,
            "forceIndexHint": self.force_index_hint,
        }


class CollectMethod(str, Enum):
    SORTED = "sorted"
    HASH = "hash"


@dataclass()
class CollectOptions(Options):
    method: CollectMethod



import logging
import sys
from typing import TYPE_CHECKING, Optional, Sequence, Type, Union, cast, overload

from pydantic import BaseModel
from pydantic.utils import lenient_issubclass

from pydango.orm.encoders import jsonable_encoder
from pydango.orm.models.base import Aliased, BaseArangoModel, LazyProxy
from pydango.orm.models.fields import ModelFieldExpression
from pydango.orm.models.utils import save_dict
from pydango.query.expressions import (
    BinaryExpression,
    CollectionExpression,
    FieldExpression,
    IteratorExpression,
    LiteralExpression,
    ReturnableExpression,
    SortExpression,
)
from pydango.query.options import (
    RemoveOptions,
    ReplaceOptions,
    UpdateOptions,
    UpsertOptions,
)
from pydango.query.query import AQLQuery, TraverseIterators

if sys.version_info >= (3, 11):
    from typing import Self, TypeAlias
else:
    from typing_extensions import Self, TypeAlias

if TYPE_CHECKING:
    from pydango.orm.models.base import ArangoModel
    from pydango.query.expressions import (
        BinaryLogicalExpression,
        ConditionExpression,
        Expression,
        IterableExpression,
        ObjectExpression,
        RangeExpression,
        VariableExpression,
    )
    from pydango.query.operations import ForParams, SortParams, TraversalDirection

logger = logging.getLogger(__name__)

ORMForParams: TypeAlias = Union["ForParams", Type[BaseArangoModel], Aliased[BaseArangoModel]]
IMPLICIT_COLLECTION_ERROR = "you must specify collection when the collection cannot be implicitly resolved"
MULTIPLE_COLLECTIONS_RESOLVED = "multiple collections resolved"


def _bind(query: "ORMQuery", node: "Expression"):
    if isinstance(node, FieldExpression):
        if node.parent and isinstance(node.parent, type) and issubclass(node.parent, BaseArangoModel):
            node.parent = query.orm_bound_vars[cast(Type[BaseArangoModel], node.parent)]

    # if isinstance(node.parent, Aliased):
    if isinstance(node, FieldExpression) and isinstance(node.parent, Aliased):
        node.parent = query.orm_bound_vars[node.parent]


def _find_models_and_bind(condition: Union["ConditionExpression", "BinaryLogicalExpression"], query: "ORMQuery"):
    stack: list[Union[BinaryExpression, "Expression"]] = [condition]
    while stack:
        current = stack.pop()
        if isinstance(current, (LiteralExpression, FieldExpression)):
            continue
        if isinstance(current, BinaryExpression):
            if isinstance(current.left, (FieldExpression, AQLQuery)):
                _bind(query, current.left)
            else:
                stack.append(current.left)

            if isinstance(current.right, (FieldExpression, AQLQuery)):
                _bind(query, current.right)
            else:
                stack.append(current.right)
        else:
            raise Exception("need to check this error")


class ORMQuery(AQLQuery):
    def __init__(self, parent: Optional[AQLQuery] = None):
        super().__init__(parent)
        # self.bind_parameter_to_sequence = {}
        self.orm_bound_vars: dict[Union[Type[BaseArangoModel], Aliased, "ModelFieldExpression"], VariableExpression] = (
            {}
        )

    def for_(
        self,
        collection_or_variable: ORMForParams,
        in_: Optional[
            Union[AQLQuery, "IterableExpression", list, "VariableExpression", list["VariableExpression"]]
        ] = None,
    ) -> Self:
        if lenient_issubclass(collection_or_variable, BaseArangoModel):
            model = cast(BaseArangoModel, collection_or_variable)
            if in_ is not None:
                raise AssertionError(
                    f"you should not pass in_ when using {collection_or_variable.__name__}"  # type: ignore[union-attr]
                )
            collection_expression = CollectionExpression(model.Collection.name)
            super().for_(collection_expression, in_)
            self.orm_bound_vars[cast(Type[BaseArangoModel], collection_or_variable)] = collection_expression.iterator

        elif isinstance(collection_or_variable, Aliased):
            if in_ is not None:
                raise AssertionError(
                    f"you should not pass in_ when using {collection_or_variable.__name__}"  # type: ignore[union-attr]
                )
            collection_expression = CollectionExpression(
                collection_or_variable.entity.Collection.name, collection_or_variable.alias
            )

            self.orm_bound_vars[collection_or_variable] = collection_expression.iterator
            super().for_(collection_expression, in_)

        elif isinstance(in_, ORMQuery):
            self.orm_bound_vars.update(in_.orm_bound_vars)
            super().for_(cast(Union[str, IteratorExpression], collection_or_variable), cast(AQLQuery, in_))

        elif isinstance(in_, LazyProxy):
            super().for_(collection_or_variable, in_.dict(by_alias=True))
        elif isinstance(in_, Sequence):
            # if in_ and isinstance(in_[0], BaseArangoModel) and len(set([x.Collection.name for x in in_])) == 1:
            #     self.bind_parameter_to_sequence[id(in_[0])] = in_[0].Collection.name
            super().for_(cast(Union[str, IteratorExpression], collection_or_variable), cast(list, in_))
        else:
            # logger.info("couldn't resolver at orm layer",extra={"vars":[collection_or_variable,in_]})
            super().for_(cast(Union[str, IteratorExpression], collection_or_variable), in_)

        return self

    def filter(self, condition: Union["ConditionExpression", "BinaryLogicalExpression"]) -> Self:
        _find_models_and_bind(condition, self)
        super().filter(condition)
        return self

    def sort(self, *sort_list: "SortParams") -> Self:
        for i in range(len(sort_list)):
            sort = sort_list[i]
            if isinstance(sort, SortExpression) and isinstance(sort.field, ModelFieldExpression):
                sort.field.parent = self.orm_bound_vars[cast(Type[BaseArangoModel], sort.field.parent)]
            if isinstance(sort, ModelFieldExpression):
                sort.parent = self.orm_bound_vars[cast(Type[BaseArangoModel], sort.parent)]

        super().sort(*sort_list)
        return self

    # noinspection PyMethodOverriding
    @overload
    def insert(self, doc: "ArangoModel") -> Self: ...

    @overload
    def insert(
        self, doc: Union[dict, "ObjectExpression", "VariableExpression"], collection: Union[str, CollectionExpression]
    ) -> Self: ...

    def insert(
        self,
        doc: Union[dict, "ObjectExpression", BaseArangoModel, "VariableExpression"],
        collection: Optional[Union[str, CollectionExpression]] = None,
    ) -> Self:
        if isinstance(doc, (BaseArangoModel, LazyProxy)):
            collection = doc.Collection.name
            doc = doc.save_dict()
        elif collection is None:
            raise ValueError(IMPLICIT_COLLECTION_ERROR)
        return super().insert(doc, collection)

    @overload
    def remove(  # noqa: PyMethodOverriding
        self, expression: BaseArangoModel, *, options: Optional[RemoveOptions] = None
    ): ...

    @overload
    def remove(
        self,
        expression: Union[dict, LiteralExpression, FieldExpression, "VariableExpression", "ObjectExpression", str],
        collection: Union[str, CollectionExpression],
        *,
        options: Optional[RemoveOptions] = None,
    ): ...

    def remove(
        self,
        expression: Union[
            BaseArangoModel, dict, LiteralExpression, FieldExpression, "VariableExpression", "ObjectExpression", str
        ],
        collection: Union[str, CollectionExpression, None] = None,
        *,
        options: Optional[RemoveOptions] = None,
    ) -> Self:
        if isinstance(expression, BaseArangoModel):
            collection = CollectionExpression(expression.Collection.name)
            expression = expression.dict(include={"key"}, by_alias=True)
        elif collection is None:
            raise ValueError(IMPLICIT_COLLECTION_ERROR)

        return super().remove(expression, collection, options=options)

    @overload
    def update(self, key, doc, *, options: Optional[UpdateOptions] = None) -> Self:  # noqa: PyMethodOverriding
        ...

    @overload
    def update(self, key, doc, coll, *, options: Optional[UpdateOptions] = None) -> Self: ...

    def update(self, key, doc, coll=None, *, options: Optional[UpdateOptions] = None) -> Self:
        if isinstance(key, BaseArangoModel):
            coll = key.Collection.name
            key = key.dict()
        if isinstance(doc, BaseModel):
            doc = doc.dict()
        super().update(key, doc, coll, options=options)
        return self

    @overload
    def replace(  # noqa: PyMethodOverriding
        self,
        key: Union[str, dict, BaseArangoModel],
        doc: BaseArangoModel,
        *,
        options: Optional[ReplaceOptions] = None,
    ) -> Self: ...

    @overload
    def replace(  # noqa: PyMethodOverriding
        self,
        key: BaseArangoModel,
        doc: Union[dict, BaseArangoModel],
        *,
        options: Optional[ReplaceOptions] = None,
    ) -> Self: ...

    @overload
    def replace(
        self,
        key: Union[str, dict, "ObjectExpression"],
        doc: Union[dict, "ObjectExpression"],
        collection: Union[str, CollectionExpression],
        *,
        options: Optional[ReplaceOptions] = None,
    ) -> Self: ...

    def replace(
        self,
        key: Union[str, dict, "ObjectExpression", BaseArangoModel],
        doc: Union[dict, "ObjectExpression", BaseArangoModel],
        collection: Union[str, CollectionExpression, None] = None,
        *,
        options: Optional[ReplaceOptions] = None,
    ) -> Self:
        if isinstance(key, BaseArangoModel):
            collection = key.Collection.name
            key = key.dict(include={"key"}, by_alias=True)
        if isinstance(doc, BaseModel):
            doc = doc.dict()
        if isinstance(doc, BaseArangoModel):
            if collection is None:
                collection = doc.Collection.name
            elif collection != doc.Collection:
                raise AssertionError(MULTIPLE_COLLECTIONS_RESOLVED)
        if collection is None:
            raise ValueError(IMPLICIT_COLLECTION_ERROR)
        return super().replace(key, doc, collection, options=options)

    @overload
    def upsert(  # noqa: PyMethodOverriding
        self,
        filter_: BaseArangoModel,
        insert: Union[dict, BaseModel, "ObjectExpression", BaseArangoModel, "VariableExpression"],
        *,
        replace: Union[dict, BaseModel, "ObjectExpression", BaseArangoModel, "VariableExpression"],
        options: Optional[UpsertOptions] = None,
    ) -> Self: ...

    @overload
    def upsert(  # noqa: PyMethodOverriding
        self,
        filter_: BaseArangoModel,
        insert: Union[dict, BaseModel, "ObjectExpression", BaseArangoModel, "VariableExpression"],
        *,
        update: Union[dict, BaseModel, "ObjectExpression", BaseArangoModel, "VariableExpression"],
        options: Optional[UpsertOptions] = None,
    ) -> Self: ...

    @overload
    def upsert(  # noqa: PyMethodOverriding
        self,
        filter_: Union[dict, BaseModel, "ObjectExpression", "VariableExpression"],
        insert: Union[dict, BaseModel, "ObjectExpression", "VariableExpression"],
        collection: Union[str, CollectionExpression],
        *,
        replace: Union[dict, BaseModel, "ObjectExpression", "VariableExpression"],
        options: Optional[UpsertOptions] = None,
    ) -> Self: ...

    @overload
    def upsert(  # noqa: PyMethodOverriding
        self,
        filter_: Union[dict, BaseModel, "ObjectExpression", "VariableExpression"],
        insert: Union[dict, BaseModel, "ObjectExpression", "VariableExpression"],
        collection: Union[str, CollectionExpression],
        *,
        update: Union[dict, BaseModel, "ObjectExpression", "VariableExpression"],
        options: Optional[UpsertOptions] = None,
    ) -> Self: ...

    def upsert(
        self,
        filter_: Union[dict, "ObjectExpression", BaseModel, BaseArangoModel, "VariableExpression"],
        insert: Union[dict, "ObjectExpression", BaseModel, BaseArangoModel, "VariableExpression"],
        collection: Union[str, CollectionExpression, None] = None,
        **kwargs,
        # update: Union[dict,"ObjectExpression", BaseModel, BaseArangoModel, None] = None,
        # replace: Union[dict, "ObjectExpression", BaseModel,BaseArangoModel, None] = None,
        # options: Optional[UpsertOptions] = None,
    ) -> Self:
        update = kwargs.get("update")
        replace = kwargs.get("replace")

        if isinstance(filter_, BaseArangoModel):
            collection = filter_.Collection.name

            if update and isinstance(update, BaseArangoModel) and update.Collection.name != collection:
                raise AssertionError(MULTIPLE_COLLECTIONS_RESOLVED)

            if replace and isinstance(replace, BaseArangoModel) and replace.Collection.name != collection:
                raise AssertionError(MULTIPLE_COLLECTIONS_RESOLVED)

        if isinstance(filter_, BaseModel):
            filter_ = filter_.dict()

        if collection is None:
            if isinstance(insert, BaseArangoModel):
                collection = insert.Collection.name
            elif update and isinstance(update, BaseArangoModel):
                collection = update.Collection.name
            elif replace and isinstance(replace, BaseArangoModel):
                collection = replace.Collection.name
            else:
                raise ValueError(IMPLICIT_COLLECTION_ERROR)

        if isinstance(insert, BaseModel):
            insert = insert.dict()
        if update and isinstance(update, BaseModel):
            kwargs["update"] = update.dict()
        if replace and isinstance(replace, BaseModel):
            kwargs["replace"] = replace.dict()

        super().upsert(filter_, insert, collection, **kwargs)
        return self

    def return_(self, return_expr: Union[Type[BaseArangoModel], Aliased, "ReturnableExpression", dict]) -> Self:
        if isinstance(return_expr, type) and issubclass(return_expr, (BaseArangoModel,)):
            return_expr = self.orm_bound_vars[return_expr]
        elif isinstance(return_expr, Aliased):
            return_expr = cast("ReturnableExpression", self.orm_bound_vars[return_expr])

        super().return_(return_expr)
        return self

    def _serialize_vars(self):
        return jsonable_encoder(self.bind_vars, by_alias=True, custom_encoder={BaseArangoModel: save_dict})

    def traverse(
        self,
        iterators: TraverseIterators,
        edges: Union[str, CollectionExpression, Sequence[Union[str, CollectionExpression]]],
        start: Union["LiteralExpression", "VariableExpression", FieldExpression, str],
        depth: Union["RangeExpression", range, tuple[int, int]],
        direction: "TraversalDirection",
    ):
        return super().traverse(iterators, edges, start, depth, direction)
        # return self


def for_(
    collection_or_variable: ORMForParams,
    in_: Optional[Union["IterableExpression", "VariableExpression", list["VariableExpression"], list]] = None,
) -> ORMQuery:
    return ORMQuery().for_(collection_or_variable, in_)


def traverse(
    collection_or_variable: ORMForParams,
    in_: Optional[Union["IterableExpression", "VariableExpression", list["VariableExpression"], list]] = None,
) -> ORMQuery:
    return ORMQuery().for_(collection_or_variable, in_)



import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Sequence, Union

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


if TYPE_CHECKING:
    from aioarango.typings import Fields


@dataclass()
class Index: ...


@dataclass()
class GeoIndex(Index):
    fields: "Fields"
    ordered: Optional[bool] = None
    name: Optional[str] = None
    in_background: Optional[bool] = None


@dataclass
class HashIndex(Index):
    fields: Sequence[str]
    unique: Optional[bool] = None
    sparse: Optional[bool] = None
    deduplicate: Optional[bool] = None
    name: Optional[str] = None
    in_background: Optional[bool] = None


@dataclass
class SkipListIndex(Index):
    fields: Sequence[str]
    unique: Optional[bool] = None
    sparse: Optional[bool] = None
    deduplicate: Optional[bool] = None
    name: Optional[str] = None
    in_background: Optional[bool] = None


@dataclass
class FullTextIndex(Index):
    fields: Sequence[str]
    min_length: Optional[int] = None
    name: Optional[str] = None
    in_background: Optional[bool] = None


@dataclass
class PersistentIndex(Index):
    fields: Sequence[str]
    unique: Optional[bool] = None
    sparse: Optional[bool] = None
    name: Optional[str] = None
    in_background: Optional[bool] = None


@dataclass
class TTLIndex(Index):
    fields: Sequence[str]
    expiry_time: int
    name: Optional[str] = None
    in_background: Optional[bool] = None


Indexes: TypeAlias = Union[GeoIndex, HashIndex, SkipListIndex, FullTextIndex, PersistentIndex, TTLIndex]



import logging
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Union, overload

from pydango.orm.encoders import jsonable_encoder
from pydango.query.expressions import (
    In,
    QueryExpression,
    SubQueryExpression,
    VectorSubQueryExpression,
)
from pydango.query.operations import (
    CollectOperation,
    FilterOperation,
    ForOperation,
    InsertOperation,
    LetOperation,
    LimitOperation,
    RemoveOperation,
    ReplaceOperation,
    ReturnOperation,
    SortOperation,
    SortParams,
    TraversalDirection,
    TraversalOperation,
    UpdateOperation,
    UpsertOperation,
)

if TYPE_CHECKING:
    from pydango.query.expressions import (
        BindableExpression,
        CollectionExpression,
        ConditionExpression,
        Expression,
        FieldExpression,
        IterableExpression,
        IteratorExpression,
        LiteralExpression,
        ObjectExpression,
        RangeExpression,
        ReturnableExpression,
        VariableExpression,
    )
    from pydango.query.operations import (
        AssignmentParam,
        AssignmentParams,
        ForParams,
        Operation,
    )
    from pydango.query.options import (
        CollectOptions,
        RemoveOptions,
        ReplaceOptions,
        UpdateOptions,
        UpsertOptions,
    )

if sys.version_info >= (3, 11):
    from typing import Self, TypeAlias
else:
    from typing_extensions import Self, TypeAlias

logger = logging.getLogger(__name__)

JsonType: TypeAlias = Union[None, int, str, bool, List["JsonType"], Dict[str, "JsonType"]]

TraverseIterators: TypeAlias = Union[
    "IteratorExpression",
    tuple["IteratorExpression"],
    tuple["IteratorExpression", "IteratorExpression"],
    tuple["IteratorExpression", "IteratorExpression", "IteratorExpression"],
]


@dataclass
class PreparedQuery:
    query: str
    bind_vars: JsonType


class AQLQuery(QueryExpression):
    sep = " "

    def __init__(self, parent: Optional["AQLQuery"] = None):
        super().__init__()
        self.compiled_vars = None
        self.bind_vars: dict[str, Union[bool, str, int, float, dict, list]] = {}
        self.__dynamic_vars__: list[VariableExpression] = []
        self.__used_vars__: set[str] = set()
        self._parameters: dict[Any, str] = {}
        self._var_counter = 0
        self._param_counter = 0
        self.parent: Optional[AQLQuery] = parent
        self._ops: list["Operation"] = []
        self._compiled = ""
        self.__is_modification_query__ = False

    def __str__(self):
        if self._compiled:
            return f"{self._compiled}{self.sep}{self.bind_vars}"
        else:
            return repr(self)

    def __repr__(self):
        aql = []
        for i in self._ops:
            if i == self:
                continue
            aql.append(repr(i))
        return self.sep.join(aql)

    def _get_var_name(self):
        if self.parent:
            return self.parent._get_var_name()
        self._var_counter += 1
        return f"var{self._var_counter}"

    def _get_param_var(self):
        if self.parent:
            return self.parent._get_param_var()
        self._param_counter += 1
        return f"@param{self._param_counter}"

    def for_(
        self,
        collection_or_variable: "ForParams",
        in_: Optional[
            Union["IterableExpression", list, "VariableExpression", list["VariableExpression"], "AQLQuery"]
        ] = None,
    ) -> Self:
        if self == in_:
            raise ValueError("is not possible to loop over the same query")

        if isinstance(in_, AQLQuery):
            in_.parent = self
            in_ = VectorSubQueryExpression(in_)

        self._ops.append(ForOperation(collection_or_variable, in_, query_ref=self))  # type: ignore[arg-type]
        return self

    def traverse(
        self,
        iterators: TraverseIterators,
        edges: Union[str, "CollectionExpression", Sequence[Union[str, "CollectionExpression"]]],
        start: Union["LiteralExpression", "VariableExpression", "FieldExpression", str],
        depth: Union["RangeExpression", range, tuple[int, int]],
        direction: TraversalDirection,
    ) -> Self:
        self._ops.append(
            TraversalOperation(
                iterators=iterators,
                edges=edges,
                start=start,
                depth=depth,
                direction=direction,
                query_ref=self,  # type: ignore[arg-type]
            )
        )
        return self

    def filter(self, condition: "ConditionExpression") -> "AQLQuery":
        if isinstance(condition, In) and isinstance(condition.right, AQLQuery):
            condition.right.parent = self
            condition.right = VectorSubQueryExpression(condition.right)

        self._ops.append(FilterOperation(condition, query_ref=self))  # type: ignore[arg-type]
        return self

    def sort(self, *sort_list: SortParams) -> "AQLQuery":
        self._ops.append(SortOperation(*sort_list, query_ref=self))  # type: ignore[arg-type]
        return self

    @overload
    def let(self, variable: "VariableExpression", expression: "Expression") -> "AQLQuery": ...

    @overload
    def let(self, variable: str, expression: "Expression") -> "VariableExpression": ...

    def let(
        self, variable: Union["VariableExpression", str], expression: "Expression"
    ) -> Union["AQLQuery", "VariableExpression"]:
        let_operation = LetOperation(variable, expression, query_ref=self)  # type: ignore[arg-type]
        self._ops.append(let_operation)
        if isinstance(variable, str):
            return let_operation.expression.variable
        return self

    def return_(self, return_expr: Union["ReturnableExpression", dict]) -> Self:
        if isinstance(return_expr, AQLQuery):
            return_expr.parent = self
            return_expr = SubQueryExpression(return_expr)
        self._ops.append(ReturnOperation(return_expr, query_ref=self))  # type: ignore[arg-type]
        return self

    def compile(self, *args, **kwargs) -> str:
        if self._compiled:
            return self._compiled
        aql = []

        for i in self._ops:
            aql.append(i.compile())

        self._compiled = self.sep.join(aql)
        return self._compiled

    def bind_variable(self) -> str:
        return self._get_var_name()

    def bind_parameter(self, parameter: "BindableExpression", override_var_name: Optional[str] = None) -> str:
        if self.parent:
            return self.parent.bind_parameter(parameter)
        is_hashable = False
        try:
            hash(parameter.value)
            is_hashable = True
            if parameter.value in self._parameters or parameter.value in self.bind_vars.keys():
                return self._parameters[parameter.value]
        except TypeError:
            if str(parameter.value) in self._parameters:
                return self._parameters[str(parameter.value)]

        var = override_var_name or self._get_param_var()

        self.bind_vars[var[1:]] = parameter.value

        if is_hashable:
            self._parameters[parameter.value] = var
        else:
            self._parameters[str(parameter.value)] = var

        return var

    def limit(self, limit) -> Self:
        self._ops.append(LimitOperation(limit, self))  # type: ignore[arg-type]
        return self

    def insert(
        self, doc: Union[dict, "ObjectExpression", "VariableExpression"], collection: Union[str, "CollectionExpression"]
    ) -> Self:
        self.__is_modification_query__ = True
        self._ops.append(InsertOperation(doc, collection, self))  # type: ignore[arg-type]
        return self

    def remove(
        self,
        expression: Union[dict, "LiteralExpression", "FieldExpression", "VariableExpression", "ObjectExpression", str],
        collection: Union[str, "CollectionExpression"],
        *,
        options: Optional["RemoveOptions"] = None,
    ) -> Self:
        self.__is_modification_query__ = True
        self._ops.append(
            RemoveOperation(expression, collection, options=options, query_ref=self)
        )  # type: ignore[arg-type]
        return self

    def update(self, key, doc, coll, *, options: Optional["UpdateOptions"] = None) -> Self:
        self.__is_modification_query__ = True
        self._ops.append(
            UpdateOperation(key, doc, coll, query_ref=self, options=options),
        )
        return self

    def replace(
        self,
        key: Union[str, dict, "ObjectExpression"],
        doc: Union["ObjectExpression", dict],
        collection: Union["CollectionExpression", str],
        *,
        options: Optional["ReplaceOptions"] = None,
    ) -> Self:
        self.__is_modification_query__ = True
        self._ops.append(
            ReplaceOperation(key, doc, collection, query_ref=self, options=options),
        )
        return self

    @overload
    def upsert(
        self,
        filter_: Union[dict, "ObjectExpression", "VariableExpression"],
        insert: Union[dict, "ObjectExpression", "VariableExpression"],
        collection: Union[str, "CollectionExpression"],
        *,
        replace: Union[dict, "ObjectExpression", "VariableExpression"],
        options: Optional["UpsertOptions"] = None,
    ) -> Self: ...

    @overload
    def upsert(
        self,
        filter_: Union[dict, "ObjectExpression", "VariableExpression"],
        insert: Union[dict, "ObjectExpression", "VariableExpression"],
        collection: Union[str, "CollectionExpression"],
        *,
        update: Union[dict, "ObjectExpression", "VariableExpression"],
        options: Optional["UpsertOptions"] = None,
    ) -> Self: ...

    def upsert(
        self,
        filter_: Union[dict, "ObjectExpression", "VariableExpression"],
        insert: Union[dict, "ObjectExpression", "VariableExpression"],
        collection: Union[str, "CollectionExpression"],
        *,
        update: Union[dict, "ObjectExpression", "VariableExpression", None] = None,
        replace: Union[dict, "ObjectExpression", "VariableExpression", None] = None,
        options: Optional["UpsertOptions"] = None,
    ) -> Self:
        self.__is_modification_query__ = True

        if update:
            upsert_operation = UpsertOperation(
                self, filter_=filter_, insert=insert, collection=collection, update=update, options=options
            )
        elif replace:
            upsert_operation = UpsertOperation(
                self, filter_=filter_, insert=insert, collection=collection, replace=replace, options=options
            )
        else:
            raise ValueError("you must pass update or replace")

        self._ops.append(upsert_operation)
        return self

    @overload
    def collect(
        self,
        *,
        collect: Optional["AssignmentParams"] = None,
        into: Optional[Union["VariableExpression", "AssignmentParam"]] = None,
        keep: Optional["VariableExpression"] = None,
        options: Optional["CollectOptions"] = None,
    ): ...

    @overload
    def collect(
        self,
        *,
        with_count_into: Optional["VariableExpression"] = None,
        options: Optional["CollectOptions"] = None,
    ): ...

    @overload
    def collect(
        self,
        *,
        collect: Optional["AssignmentParams"] = None,
        with_count_into: Optional["VariableExpression"] = None,
        options: Optional["CollectOptions"] = None,
    ): ...

    @overload
    def collect(
        self,
        *,
        aggregate: Optional["AssignmentParams"] = None,
        into: Optional[Union["VariableExpression", "AssignmentParam"]] = None,
        options: Optional["CollectOptions"] = None,
    ): ...

    @overload
    def collect(
        self,
        *,
        collect: Optional["AssignmentParams"] = None,
        aggregate: Optional["AssignmentParams"] = None,
        into: Optional[Union["VariableExpression", "AssignmentParam"]] = None,
        options: Optional["CollectOptions"] = None,
    ): ...

    def collect(
        self,
        *,
        collect: Optional["AssignmentParams"] = None,
        aggregate: Optional["AssignmentParams"] = None,
        into: Optional[Union["VariableExpression", "AssignmentParam"]] = None,
        keep: Optional["VariableExpression"] = None,
        with_count_into: Optional["VariableExpression"] = None,
        options: Optional["CollectOptions"] = None,
    ) -> Self:
        self._ops.append(
            CollectOperation(
                collect=collect,
                aggregate=aggregate,
                into=into,
                keep=keep,
                with_count_into=with_count_into,
                options=options,
                query_ref=self,
            )
        )
        return self

    def _serialize_vars(self):
        return jsonable_encoder(self.bind_vars, by_alias=True)

    def prepare(self) -> PreparedQuery:
        return PreparedQuery(self.compile(), self._serialize_vars())



import asyncio
from typing import TYPE_CHECKING, Type, cast

from aioarango.database import StandardDatabase

from pydango.connection.exceptions import SessionNotInitializedError
from pydango.connection.utils import get_or_create_collection

if TYPE_CHECKING:
    from pydango import PydangoSession
    from pydango.orm.models.base import ArangoModel


async def init_model(model: type["ArangoModel"], session: "PydangoSession"):
    if not session.initialized:
        raise SessionNotInitializedError()

    collection = await get_or_create_collection(cast(StandardDatabase, session.database), model)
    await session.create_indexes(collection, model)


async def init_models(session: "PydangoSession", *models: Type["ArangoModel"]):
    if not session.initialized:
        raise SessionNotInitializedError()
    await asyncio.gather(*[init_model(coll, session) for coll in models])



from .models import (
    EdgeCollectionConfig,
    EdgeModel,
    Relation,
    VertexCollectionConfig,
    VertexModel,
)
from .query import ORMQuery

__all__ = ["VertexModel", "EdgeModel", "EdgeCollectionConfig", "VertexCollectionConfig", "ORMQuery", "Relation"]



import asyncio
import dataclasses
import json
import logging
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    MutableMapping,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
    overload,
)

from aioarango import ArangoClient
from aioarango.collection import Collection, StandardCollection
from aioarango.database import StandardDatabase
from aioarango.exceptions import AQLQueryExecuteError
from aioarango.result import Result
from aioarango.typings import Json

from pydango.connection.consts import PYDANGO_SESSION_KEY
from pydango.connection.exceptions import (
    DocumentNotFoundError,
    SessionNotInitializedError,
)
from pydango.connection.graph_utils import db_traverse, graph_to_document
from pydango.connection.query_utils import (
    _build_graph_query,
    _get_upsert_filter,
    _make_upsert_query,
)
from pydango.connection.types import CollectionUpsertOptions, UpdateStrategy
from pydango.connection.utils import get_or_create_db
from pydango.indexes import (
    FullTextIndex,
    GeoIndex,
    HashIndex,
    Indexes,
    PersistentIndex,
    SkipListIndex,
    TTLIndex,
)
from pydango.orm.models import BaseArangoModel, VertexModel
from pydango.orm.query import ORMQuery
from pydango.query import AQLQuery
from pydango.query.expressions import IteratorExpression, VariableExpression
from pydango.query.functions import Document
from pydango.query.operations import TraversalDirection
from pydango.query.query import TraverseIterators

if TYPE_CHECKING:
    from pydango.orm.models.base import ArangoModel
    from pydango.orm.models.vertex import TVertexModel

logger = logging.getLogger(__name__)
_INDEX_MAPPING: dict[Type[Indexes], Callable[..., Awaitable["Result[Json]"]]] = {
    GeoIndex: Collection.add_geo_index,
    HashIndex: Collection.add_hash_index,
    SkipListIndex: Collection.add_skiplist_index,
    FullTextIndex: Collection.add_fulltext_index,
    PersistentIndex: Collection.add_persistent_index,
    TTLIndex: Collection.add_ttl_index,
}


def _collection_from_model(database: StandardDatabase, model: Type[BaseArangoModel]) -> StandardCollection:
    return database.collection(model.Collection.name)


class PydangoSession:
    @overload
    def __init__(self, *, database: StandardDatabase): ...

    @overload
    def __init__(
        self, *, client: ArangoClient, database: str, username: str = "", password: str = "", auth_method: str = "basic"
    ): ...

    def __init__(
        self,
        *,
        client: Optional[ArangoClient] = None,
        database: Union[StandardDatabase, str],
        username: str = "root",
        password: str = "",
        auth_method: str = "basic",
    ):
        if isinstance(database, str):
            self._db_name = database
            self.database = None
        else:
            self.database = database

        if not isinstance(database, StandardDatabase):
            self.password = password
            self.username = username
            self.client = client
            self.auth_method = auth_method

    async def initialize(self):
        if self.database is None:
            self.database = await get_or_create_db(
                self.client, self._db_name, user=self.username, password=self.password
            )

    @property
    def initialized(self):
        return isinstance(self.database, StandardDatabase)

    @staticmethod
    async def create_indexes(collection: StandardCollection, model: Type["ArangoModel"]) -> Sequence[Result[Json]]:
        if model.Collection.indexes:
            logger.debug("creating indexes", extra=dict(indexes=model.Collection.indexes, model=model))
        index_requests = []
        for i in model.Collection.indexes or []:
            if isinstance(i, dict):
                index_requests.append(_INDEX_MAPPING[i["type"]](collection, **i))
            else:
                index_requests.append(_INDEX_MAPPING[i.__class__](collection, **dataclasses.asdict(i)))

        return cast(list[Result[Json]], await asyncio.gather(*index_requests))

    async def save(
        self,
        document: "ArangoModel",
        strategy: UpdateStrategy = UpdateStrategy.UPDATE,
        # todo: follow_links: bool = False,
        collection_options: Union[CollectionUpsertOptions, None] = None,
    ) -> Union["ArangoModel", "TVertexModel"]:
        model_fields_mapping = None
        if isinstance(document, VertexModel):
            model_fields_mapping, vertices_ids, edge_ids, query = _build_graph_query(
                document, collection_options=collection_options
            )
        else:
            options = (
                collection_options
                and (collection_options.get(document.Collection.name) or collection_options.get(document.__class__))
                or None
            )

            filter_ = _get_upsert_filter(document)
            query = _make_upsert_query(filter_, document, document, ORMQuery(), strategy, options)

        try:
            cursor = await self.execute(query)
        except AQLQueryExecuteError as e:
            logger.exception(query)
            raise e
        else:
            result = await cursor.next()
        if model_fields_mapping:
            db_traverse(cast(VertexModel, document), set(), result, model_fields_mapping, vertices_ids, edge_ids)
        logger.debug("cursor stats", extra=cursor.statistics())
        return document

    async def get(
        self,
        model: Type["ArangoModel"],
        key: str,
        should_raise: bool = False,
        fetch_edges: Union[set[str], bool] = False,
        # fetch_edges_data: Union[set[str], bool] = False,
        fetch_path: bool = False,
        depth: range = range(1, 1),
        prune: bool = False,
        projection: Optional[Type["ArangoModel"]] = None,
        return_raw: bool = False,
    ) -> Optional[Union["TVertexModel", "ArangoModel"]]:
        collection = model.Collection.name
        _id = f"{collection}/{key}"
        d = Document(_id)
        doc = VariableExpression()
        main_query = ORMQuery().let(doc, d)
        return_: Union[VariableExpression, dict[str, VariableExpression]] = doc
        edges: Sequence[str]
        if fetch_edges:
            if isinstance(fetch_edges, set):
                edges = cast(Sequence[str], tuple(fetch_edges))
            else:
                _edges = []
                for i in model.__relationships__.values():
                    if i.via_model:
                        _edges.append(i.via_model.Collection.name)
                edges = _edges

            v = IteratorExpression("v")
            iterators = [v]
            e = IteratorExpression("e")
            iterators.append(e)

            if fetch_path:
                p = IteratorExpression("p")
                iterators.append(p)
            traversal_result = VariableExpression()

            traversal_iterators: TraverseIterators = cast(TraverseIterators, tuple(iterators))
            traversal = (
                ORMQuery()
                .traverse(traversal_iterators, edges, _id, depth, TraversalDirection.OUTBOUND)
                .return_({"v": iterators[0], "e": iterators[1]})
            )
            main_query.let(traversal_result, traversal)
            return_ = {"doc": doc, "edges": traversal_result}

        main_query.return_(return_)

        cursor = await self.execute(main_query)
        result = await cursor.next()
        if not result or (fetch_edges and not result.get("doc")):
            raise DocumentNotFoundError(_id)

        if issubclass(model, VertexModel):
            result, recursive = graph_to_document(result, model)

        if return_raw:
            return result

        result[PYDANGO_SESSION_KEY] = self
        if result is None and should_raise:
            raise DocumentNotFoundError()

        if projection:
            document = projection.from_orm(result, session=self)
        else:
            document = model.from_orm(result, session=self)

        return document

    async def find(self, model: Type[BaseArangoModel], filters=None, skip=None, limit=None):
        if self.database is None:
            raise SessionNotInitializedError(
                "you should call `await session.initialize()` before using the session or initialize it in the"
                " constructor with `StandardDatabase`"
            )
        collection = _collection_from_model(self.database, model)
        return await collection.find(filters, skip, limit)

    async def execute(self, query: "AQLQuery", **options):
        if self.database is None:
            raise SessionNotInitializedError(
                f"you should call `await {self.initialize.__name__}` before using the session or initialize it in the"
                " constructor with `StandardDatabase`"
            )
        prepared_query = query.prepare()
        logger.debug(
            "executing query", extra={"query": prepared_query.query, "bind_vars": json.dumps(prepared_query.bind_vars)}
        )
        return await self.database.aql.execute(
            prepared_query.query, bind_vars=cast(MutableMapping, prepared_query.bind_vars), **options
        )



from typing import TYPE_CHECKING, Awaitable, Optional, Type, Union, cast, overload

import aioarango

if TYPE_CHECKING:
    from aioarango import ArangoClient
    from aioarango.collection import StandardCollection
    from aioarango.database import StandardDatabase

    from pydango.orm.models.base import ArangoModel


@overload
async def get_or_create_collection(
    db: "StandardDatabase", model: Type["ArangoModel"], *, edge=None
) -> "StandardCollection": ...


@overload
async def get_or_create_collection(db: "StandardDatabase", model: str, *, edge=None) -> "StandardCollection": ...


async def get_or_create_collection(
    db: "StandardDatabase", model: Union[str, Type["ArangoModel"]], *, edge: Optional[bool] = None
) -> "StandardCollection":
    if isinstance(model, str):
        collection_name = model
        edge = edge or False
    elif collection := getattr(model, "Collection", None):
        collection_name = collection.name
        if edge is None:
            edge = True if collection.type.value == collection.type.EDGE else False
    else:
        raise AssertionError()

    if not await db.has_collection(collection_name):
        try:
            return await cast(Awaitable["StandardCollection"], db.create_collection(collection_name, edge=edge))
        except aioarango.exceptions.CollectionCreateError as e:
            if e.error_code != 1207:
                raise e

    return db.collection(collection_name)


async def get_or_create_db(
    client: "ArangoClient", db: str, user: str = "root", password: str = "", auth_method: str = "basic", **create_params
) -> "StandardDatabase":
    sys_db = await client.db("_system", username=user, password=password)

    if not await sys_db.has_database(db):
        await sys_db.create_database(db, **create_params)

    return await client.db(db, username=user, password=password, auth_method=auth_method)


async def deplete_cursor(cursor):
    result = []
    while cursor.has_more():  # Fetch until nothing is left on the server.
        await cursor.fetch()
    while not cursor.empty():  # Pop until nothing is left on the cursor.
        result.append(cursor.pop())
    return result


async def iterate_cursor(cursor):
    return [doc async for doc in cursor]



import json
from abc import ABC
from typing import TYPE_CHECKING, Optional, Union

from pydango.query.expressions import (
    Expression,
    IterableExpression,
    ListExpression,
    LiteralExpression,
    ObjectExpression,
    QueryExpression,
    ReturnableExpression,
)

if TYPE_CHECKING:
    from pydango.query.expressions import VariableExpression


class BaseFunctionExpression(Expression):
    def __init__(self, name, *arguments):
        self.name = name
        arguments = list(arguments)

        for i, arg in enumerate(arguments):
            if isinstance(arg, dict):
                arguments[i] = ObjectExpression(arg)
            if isinstance(arg, list):
                arguments[i] = ListExpression(arg)

        self.arguments = arguments

    def compile(self, query_ref: "QueryExpression") -> str:
        arguments = []
        for arg in self.arguments:
            if isinstance(arg, QueryExpression):
                arg.parent = query_ref

            if isinstance(arg, Expression):
                value = arg.compile(query_ref)
            elif isinstance(arg, (int, float, str, bool)):
                value = json.dumps(arg)
            else:
                value = arg
            arguments.append(value)

        return f"{self.name}({', '.join(arguments)})"

    def __repr__(self) -> str:
        arguments = [repr(arg) for arg in self.arguments]
        return f"{self.name}({', '.join(arguments)})"


class FunctionExpression(BaseFunctionExpression, ReturnableExpression):
    name: str

    def __init__(self, *arguments):
        if not hasattr(self, "name") or not isinstance(self.name, str):
            raise ValueError("function name not defined")
        super().__init__(self.name, *arguments)


# document
class Document(FunctionExpression):
    name = "DOCUMENT"

    def __init__(self, _id: str):
        super().__init__(_id)


class Unset(FunctionExpression):
    name = "UNSET"

    def __init__(self, collection, *fields):
        fields = list(fields)
        for i, field in enumerate(fields):
            if not isinstance(field, LiteralExpression):
                fields[i] = LiteralExpression(field)
        super().__init__(collection, *fields)


class Merge(
    FunctionExpression,
    ObjectExpression,
):
    name = "MERGE"

    def __init__(self, *objects):
        objects_ = []
        for o in objects:
            if isinstance(o, dict):
                objects_.append(ObjectExpression(o))
            else:
                objects_.append(o)
        super().__init__(*objects_)


class Has(FunctionExpression):
    name = "HAS"

    def __init__(self, collection: str, attr: str):
        super().__init__(collection, attr)


class Zip(FunctionExpression):
    name = "ZIP"


class Sum(FunctionExpression):
    name = "SUM"


class Length(FunctionExpression):
    name = "LENGTH"
    further = ""

    def __sub__(self, other):
        self.further = f" - {other}"
        return self

    def __add__(self, other):
        self.further = f" + {other}"
        return self

    def compile(self, query_ref: "QueryExpression") -> str:
        return super().compile(query_ref) + self.further


class CollectionsExpression(FunctionExpression):
    name = "COLLECTIONS"

    def __init__(self):
        super().__init__()


# strings
class RegExMatch(FunctionExpression):
    name = "REGEX_MATCHES"

    def __init__(
        self,
        string,
        pattern,
        case_insensitive: bool,
    ):
        super().__init__(string, pattern, case_insensitive)


# arrays
class ReturnsArray(IterableExpression, ABC):
    pass


class Append(FunctionExpression):
    name = "APPEND"

    def __init__(self, array, value):
        super().__init__(array, value)


class Concat(FunctionExpression, ReturnsArray):
    name = "CONCAT"

    def __init__(self, *arrays):
        super().__init__(*arrays)


class Count(FunctionExpression):
    name = "COUNT"

    def __init__(self, array):
        super().__init__(array)


class CountDistinct(FunctionExpression):
    name = "COUNT_DISTINCT"

    def __init__(self, array):
        super().__init__(array)


CountUnique = CountDistinct


class First(FunctionExpression):
    name = "FIRST"

    def __init__(self, array):
        super().__init__(array)


class Flatten(FunctionExpression):
    name = "FLATTEN"

    def __init__(self, array):
        super().__init__(array)


class Interleave(FunctionExpression):
    name = "INTERLEAVE"

    def __init__(self, *array):
        super().__init__(*array)


class Jaccard(FunctionExpression):
    name = "JACCARD"

    def __init__(self, array1, array2):
        super().__init__(array1, array2)


class Last(FunctionExpression):
    name = "LAST"

    def __init__(self, array):
        super().__init__(array)


class Nth(FunctionExpression):
    name = "NTH"

    def __init__(self, array, index):
        super().__init__(array, index)


class Push(FunctionExpression):
    name = "PUSH"

    def __init__(self, array, value):
        super().__init__(array, value)


class Pop(FunctionExpression):
    name = "POP"

    def __init__(self, array):
        super().__init__(array)


class RemoveNth(FunctionExpression):
    name = "REMOVE_NTH"

    def __init__(self, array, index):
        super().__init__(array, index)


class ReplaceNth(FunctionExpression):
    name = "REPLACE_NTH"

    def __init__(self, array, index):
        super().__init__(array, index)


class RemoveValue(FunctionExpression):
    name = "REMOVE_VALUE"

    def __init__(self, array, value, limit: Optional[int] = None):
        super().__init__(array, value, limit)


class Reverse(FunctionExpression):
    name = "REVERSE"

    def __init__(self, array):
        super().__init__(array)


class Shift(FunctionExpression):
    name = "SHIFT"

    def __init__(self, array):
        super().__init__(array)


class ArrayFunctionMixin:
    def __getitem__(self, item):
        return self.arguments[item]


class Slice(FunctionExpression, IterableExpression, ArrayFunctionMixin):
    name = "SLICE"

    def __init__(self, array, start, count=None):
        if count is None:
            super().__init__(array, start)
        else:
            super().__init__(array, start, count)


class UnionArrays(FunctionExpression, IterableExpression, ArrayFunctionMixin):
    name = "UNION"

    def __init__(self, *arrays: Union[ListExpression, "VariableExpression"]):
        super().__init__(*arrays)


class Difference(FunctionExpression):
    name = "DIFFERENCE"

    def __init__(self, array1, array2):
        super().__init__(array1, array2)


class Intersection(FunctionExpression):
    name = "INTERSECTION"

    def __init__(self, *arrays):
        super().__init__(*arrays)


class Outersection(FunctionExpression):
    name = "OUTERSECTION"

    def __init__(self, *array):
        super().__init__(*array)


class Minus(FunctionExpression):
    name = "MINUS"

    def __init__(self, array1, array2):
        super().__init__(array1, array2)


class Position(FunctionExpression):
    name = "POSITION"

    def __init__(self, array, value):
        super().__init__(array, value)


Contains = Position


class UnionDistinct(FunctionExpression):
    name = "UNION_DISTINCT"

    def __init__(self, *arrays):
        super().__init__(*arrays)


# bit


class BitAnd(FunctionExpression):
    name = "BIT_AND"

    def __init__(self, value1, value2):
        super().__init__(value1, value2)


class BitNot(FunctionExpression):
    name = "BIT_NOT"

    def __init__(self, value):
        super().__init__(value)


class BitOr(FunctionExpression):
    name = "BIT_OR"

    def __init__(self, value1, value2):
        super().__init__(value1, value2)


class BitXor(FunctionExpression):
    name = "BIT_XOR"

    def __init__(self, value1, value2):
        super().__init__(value1, value2)


# date


class DateAdd(FunctionExpression):
    name = "DATE_ADD"

    def __init__(self, date, amount, unit):
        super().__init__(date, amount, unit)


class DateCompare(FunctionExpression):
    name = "DATE_COMPARE"

    def __init__(self, date1, date2):
        super().__init__(date1, date2)


class DateDay(FunctionExpression):
    name = "DATE_DAY"

    def __init__(self, date):
        super().__init__(date)


class DateDayOfWeek(FunctionExpression):
    name = "DATE_DAY_OF_WEEK"

    def __init__(self, date):
        super().__init__(date)


class DateDaysInMonth(FunctionExpression):
    name = "DATE_DAYS_IN_MONTH"

    def __init__(self, date):
        super().__init__(date)


class DateDiff(FunctionExpression):
    name = "DATE_DIFF"

    def __init__(self, date1, date2, unit):
        super().__init__(date1, date2, unit)


class DateHour(FunctionExpression):
    name = "DATE_HOUR"

    def __init__(self, date):
        super().__init__(date)


class DateMilliseconds(FunctionExpression):
    name = "DATE_MILLISECONDS"

    def __init__(self, date):
        super().__init__(date)


class DateMinute(FunctionExpression):
    name = "DATE_MINUTE"

    def __init__(self, date):
        super().__init__(date)


class DateMonth(FunctionExpression):
    name = "DATE_MONTH"

    def __init__(self, date):
        super().__init__(date)


class DateNow(FunctionExpression):
    name = "DATE_NOW"

    def __init__(self):
        super().__init__()


class DateSecond(FunctionExpression):
    name = "DATE_SECOND"

    def __init__(self, date):
        super().__init__(date)


class DateSubtract(FunctionExpression):
    name = "DATE_SUBTRACT"

    def __init__(self, date, amount, unit):
        super().__init__(date, amount, unit)


class DateTimestamp(FunctionExpression):
    name = "DATE_TIMESTAMP"

    def __init__(self, date):
        super().__init__(date)


class DateYear(FunctionExpression):
    name = "DATE_YEAR"

    def __init__(self, date):
        super().__init__(date)



from typing import Optional, Sequence, Union

from aioarango import ArangoClient, HTTPClient
from aioarango.connection import (
    BasicConnection,
    Connection,
    JwtConnection,
    JwtSuperuserConnection,
)


def make_connection(  # nosec: B107
    hosts: Union[str, Sequence[str]] = "http://127.0.0.1:8529",
    host_resolver: str = "roundrobin",
    http_client: Optional[HTTPClient] = None,
    db_name: str = "_system",
    username: str = "root",
    password: str = "",
    auth_method: str = "basic",
    superuser_token: Optional[str] = None,
) -> Connection:
    client = ArangoClient(hosts, host_resolver, http_client)

    if superuser_token is not None:
        return JwtSuperuserConnection(
            hosts=client.hosts,
            host_resolver=client._host_resolver,
            sessions=client._sessions,
            db_name=db_name,
            http_client=client._http,
            serializer=client._serializer,
            deserializer=client._deserializer,
            superuser_token=superuser_token,
        )

    elif auth_method.lower() == "basic":
        return BasicConnection(
            hosts=client.hosts,
            host_resolver=client._host_resolver,
            sessions=client._sessions,
            db_name=db_name,
            username=username,
            password=password,
            http_client=client._http,
            serializer=client._serializer,
            deserializer=client._deserializer,
        )
    elif auth_method.lower() == "jwt":
        return JwtConnection(
            hosts=client.hosts,
            host_resolver=client._host_resolver,
            sessions=client._sessions,
            db_name=db_name,
            username=username,
            password=password,
            http_client=client._http,
            serializer=client._serializer,
            deserializer=client._deserializer,
        )

    else:
        raise ValueError(f"invalid auth_method: {auth_method}")



from typing import Protocol, TypeVar, Union

from pydango.query.consts import FROM, ID, KEY, REV, TO
from pydango.query.expressions import NEW


class Compilable(Protocol):
    def compile(self, *args, **kwargs) -> Union[str, None]: ...


T = TypeVar("T")


def new(*, edge=False, debug=True) -> Union[dict[str, str], NEW]:
    _new = NEW()
    if debug:
        return _new
    d = {ID: _new[ID], KEY: _new[KEY], REV: _new[REV]}
    if edge:
        d.update({FROM: _new[FROM], TO: _new[TO]})
    return d
